(*
 This file is part of Poussin.

 Poussin is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Poussin is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Poussin.  If not, see <http://www.gnu.org/licenses/>.

 Copyright (C) 2008 Nicolas Marti
*)

open List;;
open String;;
open Printf;;

open Def;;
open Pprinter;;
open Listutil;;
open Freshness;;
open Term;;
open Varset;;
open Betared;;
open Unfold;;

(* a term equivalence *)

let rec alpha_rel' t1 t2 l def =
  (*
    printf "alpha_rel'\n";
    printf "t1 = "; print_term (listfold_term_term_var_fp t1 def); printf "\n";
    printf "t2 = "; print_term (listfold_term_term_var_fp t2 def); printf "\n\n";
    printf "muri(alpha_rel') \n\n";
  *)  
  match (t1, t2) with

    | (Var x1, Var x2) ->
        if (x1 = x2) then (true, l) else (

          (*printf "alpha_rel(var): %s =/= %s\n\n" x1 x2;*)
          (false, l)

        )

    | (Type, Type) -> (true, l)

    | (Lambda (x1, t11,t12), Lambda (x2, t21, t22)) ->
        if (x1 = x2) then (
          match (alpha_rel' t11 t21 (x1 ++ l) def) with
            | (true, l1) -> (
                match (alpha_rel' t12 t22 l1 def) with
                  | (true, l2) ->
                      (true, l2)
                  | (false, l2) ->
                      ((false, l2))
              )
            | (false, l1) -> ((false, l1))
        ) else (
          let x = (fresh_name_list_name (l +++ (term_set_var (Lambda (x1, t11,t12))) +++ (term_set_var (Lambda (x2, t21,t22)))) x1) in
          match (cf_term' (rewrite_term_var_term t11 x1 (Var x)) (x ++ l), cf_term' (rewrite_term_var_term t21 x2 (Var x)) (x ++ l)) with
            | ((t11', l1'),(t21', l2')) ->
                match (cf_term' (rewrite_term_var_term t12 x1 (Var x)) l1', cf_term' (rewrite_term_var_term t22 x2 (Var x)) l2') with
                  | ((t12', l1''),(t22', l2'')) ->
                      match (alpha_rel' t11' t21' (x ++ l) def) with
                        | (true, l1) -> (
                            match (alpha_rel' t12' t22' l1 def) with
                              | (true, l2) ->
                                  (true, l2)
                              | (false, l2) ->
                                  ((false, l2))
                          )
                        | (false, l1) -> ((false, l1))
                            
        ) 

    | (Forall (x1, t11,t12), Forall (x2, t21, t22)) ->
        if (x1 = x2) then (
          match (alpha_rel' t11 t21 (x1 ++ l) def) with
            | (true, l1) -> (
                match (alpha_rel' t12 t22 l1 def) with
                  | (true, l2) ->
                      (true, l2)
                  | (false, l2) ->
                      ((false, l2))
              )
            | (false, l1) -> ((false, l1))
        ) else (
          let x = (fresh_name_list_name (l +++ (term_set_var (Forall (x1, t11,t12))) +++ (term_set_var (Forall (x2, t21,t22)))) x1) in
          match (cf_term' (rewrite_term_var_term t11 x1 (Var x)) (x ++ l), cf_term' (rewrite_term_var_term t21 x2 (Var x)) (x ++ l)) with
            | ((t11', l1'),(t21', l2')) ->
                match (cf_term' (rewrite_term_var_term t12 x1 (Var x)) l1', cf_term' (rewrite_term_var_term t22 x2 (Var x)) l2') with
                  | ((t12', l1''),(t22', l2'')) ->
                      match (alpha_rel' t11' t21' (x ++ l) def) with
                        | (true, l1) -> (
                            match (alpha_rel' t12' t22' l1 def) with
                              | (true, l2) ->
                                  (true, l2)
                              | (false, l2) ->
                                  (false, l2)
                          )
                        | (false, l1) -> (false, l1)               
                            
        ) 

    | (Phi (x1, n1, t11,t12), Phi (x2, n2, t21, t22)) ->
        if (x1 = x2) then (
          let (t11_concl, t11_lcv) = (forall_concl_list_var_type t11) in
          let (t21_concl, t21_lcv) = (forall_concl_list_var_type t21) in
            alpha_rel' 
              
              (concl_list_var_type_lambda
                  (App (t11_concl, t12))
                  t11_lcv
              )

              (concl_list_var_type_lambda
                  (App (t21_concl, t22))
                  t21_lcv
              )
              
              (x1 ++ (x2 ++ l)) def
            
            
        ) else (
          (* Here is the probleme !!!! *)
          let l' = (l +++ (term_set_var (Phi (x1, n1, t11,t12))) +++ (term_set_var (Phi (x2, n2, t21,t22)))) in
          let x = (fresh_name_list_name l' x1) in
          let t12' = (rewrite_term_var_term t12 x1 (Var x)) in
          let t22' = (rewrite_term_var_term t22 x2 (Var x)) in
          let (t11_concl, t11_lcv) = (forall_concl_list_var_type t11) in
          let (t21_concl, t21_lcv) = (forall_concl_list_var_type t21) in
            match (alpha_rel' 
                      
                      (concl_list_var_type_lambda
                          (App (t11_concl, t12'))
                          t11_lcv
                      )
                      
                      (concl_list_var_type_lambda
                          (App (t21_concl, t22'))
                          t21_lcv
                      )
                      
                      (x1 ++ (x ++ (x2 ++ l'))) def
            ) with
              | (false, l'') ->
                  (false, l'')
              | t'' -> t''
                            
        ) 

    | (Ind (x1, t11,t12), Ind (x2, t21, t22)) ->
        if (x1 = x2) then (
          match (alpha_rel' t11 t21 (x1 ++ l) def) with
            | (true, l1) -> (
                match (alpha_rel_list' t12 t22 l1 def) with
                  | (true, l2) ->
                      (true, l2)
                  | (false, l2) ->
                      ((false, l2))
              )
            | (false, l1) -> ((false, l1))
        ) else (
          let x = (fresh_name_list_name (l +++ (term_set_var (Ind (x1, t11,t12))) +++ (term_set_var (Ind (x2, t21,t22)))) x1) in
          match (cf_term' (rewrite_term_var_term t11 x1 (Var x)) (x ++ l), cf_term' (rewrite_term_var_term t21 x2 (Var x)) (x ++ l)) with
            | ((t11', l1'),(t21', l2')) ->
                match (cf_listterm' (rewrite_listterm_var_term t12 x1 (Var x)) l1', cf_listterm' (rewrite_listterm_var_term t22 x2 (Var x)) l2') with
                  | ((t12', l1''),(t22', l2'')) ->
                      match (alpha_rel' t11' t21' (x ++ l) def) with
                        | (true, l1) -> (
                            match (alpha_rel_list' t12' t22' l1 def) with
                              | (true, l2) ->
                                  (true, l2)
                              | (false, l2) ->
                                  ((false, l2))
                          )
                        | (false, l1) -> ((false, l1))
                            
        ) 

    | (Cons (n1, t11), Cons (n2, t21)) ->
        if (n1 = n2) then (alpha_rel' t11 t21 l def) else (false, l)

    | (App (t11, t12), App (t21, t22)) -> (
        match (alpha_rel' t11 t21 l def) with
          | (true, l1) -> (
              match (alpha_rel' t12 t22 l1 def) with
                | (true, l2) ->
                    (true, l2)
                | (false, l2) ->
                    ((*printf "alpha_rel(app(1))\n\n";*) false, l2)
            )
          | (false, l1) -> 
              (*(((*printf "alpha_rel(app(2))\n\n"; *) false, l1))*)
              (beta_red_alpha_rel' t1 t2 l def)
      )

    | (Match (t11, t12), Match (t21, t22)) -> (
        match (alpha_rel' t11 t21 l def) with
          | (true, l1) -> (
              match (alpha_rel_list' t12 t22 l1 def) with
                | (true, l2) ->
                    (true, l2)
                | (false, l2) ->
                    ((false, l2))
            )
          | (false, l1) -> ((false, l1))
      )

    | (MatchT (t11, t12, t13), MatchT (t21, t22, t23)) -> (
        
        (alpha_rel' (Match (t11, t12)) (Match (t21, t22)) l def)
        
      )

    | _ -> (
        (*
        printf "alpha_rel'\n";
        printf "t1 = "; print_term (listfold_term_term_var_fp t1 def); printf "\n";
        printf "t2 = "; print_term (listfold_term_term_var_fp t2 def); printf "\n\n";
        printf "muri(alpha_rel') \n\n";
        *)
        (beta_red_alpha_rel' t1 t2 l def)
                
      )

and alpha_rel_list' t1 t2 l def =  
  (*printf "alpha_rel_list'\n\n"; *)
  match (t1, t2) with
    | ([], []) -> 
        ((true, l))
    | (hd1::tl1, hd2::tl2) -> (
        match (alpha_rel' hd1 hd2 l def) with
          | (true, l1) ->              
              (alpha_rel_list' tl1 tl2 l def)
          | (false, l1) -> (false, l1)
      )
    | _ -> ((*printf "muri(alpha_rel_list')\n";*) (false, l))
and beta_red_alpha_rel' t1 t2 l def =
  let t1' = (beta_red (listrewrite_term_var_term t1 def) l []) in
  let t2' = (beta_red (listrewrite_term_var_term t2 def) l []) in
    if (t1' = t1 && t2 = t2') then (false, l) else
      alpha_rel' t1' t2' l def
;;

let alpha_rel t1 t2 def = 
  let (b, l) = (alpha_rel' t1 t2 VarSet.empty def) in b;;

