(*
 This file is part of Poussin.

 Poussin is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Poussin is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Poussin.  If not, see <http://www.gnu.org/licenses/>.

 Copyright (C) 2008 Nicolas Marti
*)

open String;;
open Printf;;

(*

  This is bullshit code: data structures do not feet, and the decision is far from complete

*)

type nexpr =
    | Nvar of string
    | Nval of int
    | Nplus of nexpr * nexpr
;;

type bexpr =
    | True 
    | False
    | Eq of nexpr * nexpr
    | Gt of nexpr * nexpr
    | And of bexpr * bexpr
    | Impl of bexpr * bexpr
;;


let rec print_nexpr n =
  match n with
    | Nvar v -> printf "%s" v;
    | Nval i -> printf "%d" i;
    | Nplus (n1, n2) -> print_nexpr n1; printf " + "; print_nexpr n2;
;;


let rec print_bexpr b =
  match b with
    | True -> printf "true";
    | False -> printf "false";
    | Eq (n1, n2) -> print_nexpr n1; printf " == "; print_nexpr n2; 
    | Gt (n1, n2) -> print_nexpr n1; printf " >> "; print_nexpr n2; 
    | And (b1, b2) -> printf "("; print_bexpr b1; printf " /\\ "; print_bexpr b2; printf ")"; 
    | Impl (b1, b2) -> printf "("; print_bexpr b1; printf " -> "; print_bexpr b2; printf ")"; 
;;

type bexpr2 =
  | True2 
  | False2
  | Eq2 of nexpr * nexpr
  | Gt2 of nexpr * nexpr
;;

let rec add_hyp l b =
  match l with
    | [] -> []
    | (hd1, hd2) ::tl ->
        (hd1 @ b, hd2) :: (add_hyp tl b);;

let rec add_con l b =
  match l with
    | [] -> []
    | (hd1, hd2) ::tl ->
        (hd1, hd2 @ b) :: (add_con tl b);;



let rec bexpr_to_bexpr2_hyp b =
  match b with
    | True -> []
    | False -> (False2::[])
    | Eq (n1, n2) -> ((Eq2 (n1, n2))::[])
    | Gt (n1, n2) -> ((Gt2 (n1, n2))::[])
    | And (b1, b2) -> (bexpr_to_bexpr2_hyp b1) @ (bexpr_to_bexpr2_hyp b2)
    | _ -> (True2::[])
;;
   
let rec bexpr_to_bexpr2 b l =
  match b with
    | True -> []
    | False -> ([],False2::[])::[]
    | Eq (n1, n2) -> add_con l (Eq2(n1,n2)::[])
    | Gt (n1, n2) -> add_con l (Gt2(n1,n2)::[])
    | And (b1, b2) -> (bexpr_to_bexpr2 b1 l) @ (bexpr_to_bexpr2 b2 l)
    | Impl (b1, b2) -> bexpr_to_bexpr2 b2 (add_hyp l (bexpr_to_bexpr2_hyp b1))
;;

let print_bexpr2 b =
  match b with
    | True2 -> printf "true";
    | False2 -> printf "false";
    | Eq2 (n1, n2) -> print_nexpr n1; printf " == "; print_nexpr n2; 
    | Gt2 (n1, n2) -> print_nexpr n1; printf " >> "; print_nexpr n2; 
;;

let rec conj_bexpr2 l =
  match l with
    | [] -> printf "";
    | hd :: tl ->
        print_bexpr2 hd;
        match tl with
          | [] -> printf "";
          | _ -> printf " /\\ "; conj_bexpr2 tl;
;;

let form2 f = 
  let (b1, b2) = f in
    conj_bexpr2 b1; printf " ->\n"; conj_bexpr2 b2;;

let rec list_conj_bexpr2 l =
  match l with
    | [] -> printf "";
    | hd :: tl ->
        conj_bexpr2 hd; printf "\n\n"; list_conj_bexpr2 tl;;

let rec list_form2 l =
  match l with
    | [] -> printf "";
    | hd :: tl ->
        form2 hd; printf "\n\n"; list_form2 tl;;

let rec rewrite_nexpr b x t =
  match b with
    | Nvar v -> if (x = v) then t else Nvar v
    | Nval v -> Nval v
    | Nplus (n1, n2) -> Nplus (rewrite_nexpr n1 x t, rewrite_nexpr n2 x t)
;;

let rec rewrite_bexpr2 b x t =
  match b with
    | True2 -> True2
    | False2 -> False2
    | Eq2 (n1, n2) -> Eq2 (rewrite_nexpr n1 x t, rewrite_nexpr n2 x t)
    | Gt2 (n1, n2) -> Gt2 (rewrite_nexpr n1 x t, rewrite_nexpr n2 x t);;

let rec rewrite_list_bexpr2 l x t =
  match l with
    | [] -> []
    | hd :: tl ->
        (rewrite_bexpr2 hd x t) :: (rewrite_list_bexpr2 tl x t);;

let rec resolve_hyp_bexpr2 h c =
  match h with
    | [] -> c
    | hd :: tl ->
        match hd with          
          | False2 -> True2::[]
          | Eq2 (Nvar x, n2) -> resolve_hyp_bexpr2 tl (rewrite_list_bexpr2 c x n2)
          | _ -> resolve_hyp_bexpr2 tl c;;

let rec resolve_list_hyp_bexpr2 l =
  match l with
    | [] -> []
    | (hd1, hd2) :: tl ->
        (resolve_hyp_bexpr2 hd1 hd2) :: (resolve_list_hyp_bexpr2 tl);;

let rec remove_var_nexpr n x =
  match n with
    | Nvar v -> if (x = v) then Some (Nval 0) else None
    | Nval i -> None
    | Nplus (n1, n2) -> 
        match (remove_var_nexpr n1 x) with
          | None -> (
              match (remove_var_nexpr n2 x) with
                | None -> None
                | Some n2' -> Some (Nplus (n1, n2'))
            )
          | Some n1' ->
              Some (Nplus (n1', n2))
;;

let rec list_var_nexpr n =
  match n with
    | Nvar v -> v :: []
    | Nval i -> []
    | Nplus (n1, n2) -> (list_var_nexpr n1) @ (list_var_nexpr n2)
;;

let rec list_var_bexpr2 b =
  match b with
    | True2 -> []
    | False2 -> []
    | Eq2 (n1, n2) -> (list_var_nexpr n1) @ (list_var_nexpr n2)
    | Gt2 (n1, n2) -> (list_var_nexpr n1) @ (list_var_nexpr n2)
;;

let rec remove_list_var_Gt_nexprs n1 n2 l =
  match l with
    | [] -> Some (n1, n2)
    | hd :: tl ->
        match (remove_var_nexpr n1 hd, remove_var_nexpr n2 hd) with
          | (Some n1', Some n2') -> remove_list_var_Gt_nexprs n1' n2' tl
          | _ -> None
;;

(* n1 >> n2 *)

let elim_var_gt n1 n2 =
  let l = (list_var_nexpr n2) in
    match (remove_list_var_Gt_nexprs n1 n2 l) with
      | None -> False2
      | Some (n1', n2') -> Gt2(n1', n2')
;;

let rec elim_var_bexpr b =
  match b with
    | True2 -> True2
    | False2 -> False2
    | Eq2 (n1, n2) -> False2
    | Gt2 (n1, n2) -> elim_var_gt n1 n2
;;

let rec elim_var_listbexpr l =
  match l with
    | [] -> []
    | hd :: tl ->
        (elim_var_bexpr hd) :: (elim_var_listbexpr tl);;

let rec list_elim_var_listbexpr l =
  match l with
    | [] -> []
    | hd :: tl ->
        (elim_var_listbexpr hd) :: (list_elim_var_listbexpr tl);;

let rec compute_lb_nexpr n =
  match n with
    | Nvar x -> 0
    | Nval i -> i
    | Nplus (n1, n2) -> (compute_lb_nexpr n1) + (compute_lb_nexpr n2);;

let rec compute_ub_nexpr n =
  match n with
    | Nvar x -> None
    | Nval i -> Some i
    | Nplus (n1, n2) -> 
        match (compute_ub_nexpr n1, compute_ub_nexpr n2) with
          | (Some i1, Some i2) -> Some (i1 + i2)
          | _ -> None
;;

let val_gt n1 n2 =
  match (compute_ub_nexpr n2) with
    | None -> False2
    | Some i2 ->
        if (compute_lb_nexpr n1 > i2) then
          True2 else False2;;

let val_bexpr2 b =
  match b with
    | True2 -> True2
    | False2 -> False2
    | Eq2 (n1, n2) -> False2
    | Gt2 (n1, n2) -> val_gt n1 n2
;;

let rec val_listbexpr2 l =
  match l with
    | [] -> []
    | hd :: tl ->
        (val_bexpr2 hd) :: (val_listbexpr2 tl);;

let rec list_val_listbexpr2 l =
  match l with
    | [] -> []
    | hd :: tl ->
        (val_listbexpr2 hd) :: (list_val_listbexpr2 tl);;


let eval_bexpr2 b =
  match b with
    | True2 -> true
    | _ -> false
;;

let rec eval_listbexpr2 l =
  match l with
    | [] -> true
    | hd :: tl ->
        match (eval_bexpr2 hd) with
          | true -> eval_listbexpr2 tl
          | false -> false;;

let rec list_eval_listbexpr2 l =
  match l with 
    | [] -> true
    | hd :: tl ->
        match (eval_listbexpr2 hd) with
          | true -> list_eval_listbexpr2 tl
          | false -> false;;


(*

(((l == 1 + A0 + 0 /\ (A0 >> 0 /\ true)) -> true) /\ (((l == 1 + A0 + e + l1 + 0 /\ (A0 >> 0 /\ (e >> 0 /\ (l1 >> 0 /\ true)))) -> (true /\ l + 0 >> l1 + 0)) /\ true))

*)
