(*
 This file is part of Poussin.

 Poussin is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Pousin is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Poussin.  If not, see <http://www.gnu.org/licenses/>.

 Copyright (C) 2008 Nicolas Marti
*)

open List;;
open String;;
open Varset;;

(* returns the nth element of a list *)
let rec nth l n =
  match l with
    | [] -> None
    | hd :: tl ->
        match n with
          | 0 -> Some hd
          | _ -> (nth tl (n - 1));;

(* a name is inside a list of name *)
let rec in_list l n =
  match l with
    | [] -> false
    | hd :: tl -> if (n = hd) then true else (in_list tl n);;

(* zip has a utility ??? :D *)

let rec zip l1 l2 =
  match (l1, l2) with
    | (hd1::tl1,hd2::tl2) ->
        (hd1,hd2)::(zip tl1 tl2)
    | _ -> [];;



(* build a list of length n with some element x *)
let rec built_n_elt_list e n =
  match n with
    | 0 -> []
    | _ -> e::(built_n_elt_list e (n - 1));;


(* keep the n first elements of the list l *)
let rec headth l n =
  match n with
    | 0 -> []
    | _ -> 
        match l with
          | [] -> []
          | hd :: tl ->
              hd :: (headth tl (n - 1));;

(* return the first projections of a list of couples *)
let rec list_proj1 l =
  match l with
    | [] -> []
    | (hd1, hd2) :: tl ->
        hd1 :: (list_proj1 tl);;

let rec list_proj1_set l =
  match l with
    | [] -> VarSet.empty
    | (hd1, hd2) :: tl ->
        hd1 ++ (list_proj1_set tl);;


let rec list_proj2 l =
  match l with
    | [] -> []
    | (hd1, hd2) :: tl ->
        hd2 :: (list_proj2 tl);;


let rec inter l1 l2 = 
  match l1 with
    | [] -> []
    | hd :: tl ->
        (if (in_list l2 hd) then (hd::[]) else []) @ (inter tl l2);;
