%{

(*
 This file is part of Poussin.

 Poussin is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Poussin is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Poussin.  If not, see <http://www.gnu.org/licenses/>.

 Copyright (C) 2008 Nicolas Marti
*)


  open Printf;;
  open Lexing;;
  open List;;
  open String;;
  
  open Def;;
  open Pprinter;;
  open Listutil;;
  open Freshness;;
  open Term;;
  open Alpharel;;
  open Betared;;
  open Termequiv;;
  open Typechecking
  open Unfold;;
  open Proofbuild;;
  open Varset;;
  open Unification;;
  open Rewriting;;
  open Kernel;;

  open Tactics;;

  open Arith;;
  open Termination;;

  let print_subgoal ctxt ty =
    printf "\n\n";
    print_subgoal_context (readable_context_fp ctxt state.def);
    printf "----------------------------\n\n";
    print_term (readable_term_fp ty state.def); printf "\n\n";;          


  let rec print_subgoals sg =
    match sg with
      | [] -> printf "\n\n";
      | (hd1, hd2, hd3) :: tl ->
          printf "----------------------------\n";
          (*print_term (readable_term_fp hd3 state.def); printf "\n\n";*)
          print_term hd3; printf "\n\n";
          print_subgoals tl;;

  let print_subgoals_ctxt () =
    match (state.pt) with
      | None -> printf "Error: Not in proof mode\n\n";
      | Some ((hd1,hd2,hd3)::tl,(ln,te,ty)) ->  (

          printf "\n\n";
          (*print_subgoal_context (readable_context_fp hd1 state.def);*)
          print_subgoal_context hd1;
          print_subgoals ((hd1,hd2,hd3)::tl);
          flush stdout;
        
        );
      | Some ([], g) -> printf "No more subgoals\n\n";;

  let print_first_subgoal () =
    match (state.pt) with
      | None -> printf "Error: Not in proof mode\n\n";
      | Some ((hd1,hd2,hd3)::tl,(ln,te,ty)) ->  (
          
            print_subgoal hd1 hd3
        
        );
      | Some ([], g) -> printf "No more subgoals\n\n";;


  let rec beta_red_ctxt c x =
    match c with 
      | [] -> []
      | (hd1, hd2) :: tl ->
          if (hd1 = x) then (hd1, beta_red hd2 VarSet.empty []) :: tl else
            (hd1, hd2) :: (beta_red_ctxt tl x);;

  let rec equiv_ctxt te ctxt name =
    match ctxt with
      | [] -> []
      | (hd1, hd2) :: tl ->
          if (hd1 = name) then (
            
            if (term_equiv te hd2 (context_to_setvar ctxt) []) then             
              (hd1, te)::tl else (hd1, hd2)::tl
                
          ) else (hd1, hd2) :: (equiv_ctxt te tl name);;
            

  let rec cf_def def bv =
    match def with
      | [] -> []
      | (hd1, hd2) :: tl -> (hd1, cf_term_listvar hd2 bv) :: (cf_def tl bv);;

  let rec cf_listdef ldef bv =
    match ldef with
      | [] -> []
      | hd :: tl -> (cf_def hd bv) :: (cf_listdef tl bv);;
    

%}

/*  */
%token NEWLINE EOF

/* The indentifiers variables */
%token <string> NAME

/* Integer (for indexes of contructors) */
%token <int> NUM
  
/* quantifier */
%token LAMBDA IND PHI FORALL
  
/* keywords */
%token MATCH RETURN WITH CONS OF TYPE INDUCTIVE DEFINITION RECURSIVE LEMMA QUIT COMPUTE DROP QED EXTRACTDEF EQUIV INTRO GENERALIZE AS APPLY EXACT SIMPL IN FACTORIZE DESTRUCT DEFINED UNFOLD
  
/* parenthese */
%token LPAREN RPAREN LBRACK RBRACK LACCOL RACCOL

/* ponctuation */
%token SCOMMA COMMA DOT DDOT TURNSTIL EQ PAR
  
/* kernel functions */
%token CHECK PRINT SHOWPROOF SHOWGOAL

%start input command
%type <unit> input command

%% /* Grammar rules and actions follow */

  input:    

| command input {}

| EOF {}
  ;               

  command :

/***********************/
/* Definition commands */
/***********************/

| INDUCTIVE NAME DDOT term DDOT EQ constructors DOT {

    if (proof_mode()) then (

      printf "Error: actually in proof mode\n\n";

    ) else (

      add_inductive_woa $2 $4 $7

    );


  }



| INDUCTIVE NAME listargs DDOT term DDOT EQ constructors DOT {

    if (proof_mode()) then (

      printf "Error:  actually in proof mode\n\n";

    ) else (

      add_inductive_wa $2 $3 $5 $8

    );

  }

| INDUCTIVE NAME DDOT term DDOT EQ DOT {

    if (proof_mode()) then (

      printf "Error: actually in proof mode\n\n";

    ) else (

      add_inductive_woa $2 $4 []

    );


  }

| INDUCTIVE NAME listargs DDOT term DDOT EQ DOT {

    if (proof_mode()) then (

      printf "Error:  actually in proof mode\n\n";

    ) else (

      add_inductive_wa $2 $3 $5 []

    );

  }


| RECURSIVE NAME listargs LBRACK listnum RBRACK DDOT term DDOT EQ term DOT {

    if (proof_mode()) then (

      printf "Error: actually in proof mode\n\n";

    ) else (

      let ctxt = ((rev (listrewrite_context_var_term $3 (listdef2def state.def))) @ state.ctxt) in
      let t = (listrewrite_term_var_term $11 (listdef2def state.def)) in
      let args = (list_name2list_var (list_proj1 $3)) in
      let indexes = $5 in
      let recname = $2 in        
      let f = term_build_term_form t ctxt args indexes recname (listdef2def state.def) in

        let f2 = bexpr_to_bexpr2 f (([],[])::[]) in

          let f3 = resolve_list_hyp_bexpr2 f2 in

            let f4 = list_elim_var_listbexpr f3 in

              let f5 = list_val_listbexpr2 f4 in

                let b = list_eval_listbexpr2 f5 in

                  if b then 
                    add_recursive $2 $3 $8 $11
                  else (
                    printf "Function %s does not seemed to finish ...\n\n" $2;
                    printf "formula:\n"; print_bexpr f; printf "\n\n";                    
                    printf "formula 2:\n"; list_form2 f2; printf "\n\n";               
                    printf "formula 3:\n"; list_conj_bexpr2 f3; printf "\n\n";
                    printf "formula 4:\n"; list_conj_bexpr2 f4; printf "\n\n";
                    printf "formula 4:\n"; list_conj_bexpr2 f5; printf "\n\n";
                    printf "result: %b \n\n" b;
                  )
    );


  }


| DEFINITION NAME DDOT term DDOT EQ term DOT {

    if (proof_mode()) then (

      printf "Error: actually in proof mode\n\n";

    ) else (

      add_definition_woa $2 $4 $7

    );

      
  }

| DEFINITION NAME listargs DDOT term DDOT EQ term DOT {

    if (proof_mode()) then (

      printf "Error: actually in proof mode\n\n";

    ) else (

      add_definition_wa $2 $3 $5 $8

    );
      
  }

/***********************/
/* Vernacular commands */
/***********************/



| COMPUTE term DOT {

    print_term (readable_term_fp (exec (listrewrite_term_var_term $2 (listdef2def state.def)) (listdef2def state.def) VarSet.empty) (state.def)); flush stdout; printf "\n\n";

  }



| CHECK term DOT {

    match (state.pt) with
      | Some ((hd1,hd2,hd3)::tl,g) ->  (

          match (
            typeinfer2
              (hd1 @ state.ctxt)
              $2
              (listdef2def state.def)
              (context_to_setvar (hd1 @ state.ctxt))
          ) with
            | None -> printf "Error: Cannot infer!\n\n"; 
            | Some ty -> (
                
                print_term $2; printf " : "; print_term (readable_term_fp ty state.def); printf "\n\n";
                
              )

        );
      | _ ->

          match (
            typeinfer2
              state.ctxt
              $2
              (listdef2def state.def)
              (context_to_setvar state.ctxt)
          ) with
            | None -> printf "Error: Cannot infer!\n\n"
            | Some ty -> (
                
                print_term $2; printf " : "; print_term (readable_term_fp ty state.def); printf "\n\n";
                
              )

    

  }


| DROP DOT {

    state.pt <- None;
    state.subgoals <- 0;

  }


| EXTRACTDEF DOT {

    printf "------------------------------------\n";
    print_def (rev (listdef2def state.def)); 
    printf "------------------------------------\n";
    print_def (rev state.lemmas_def);
    printf "------------------------------------\n";
    print_def (state.ctxt);
    printf "------------------------------------\n\n";
    
  }

| SHOWPROOF DOT {

    match (state.pt) with
      | None -> printf "Error: Not in proof mode\n\n";
      | Some (l,(n,te,ty)) ->  
          print_context state.ctxt; printf " |- \n";
          print_term te; printf " : ";
          print_term ty; printf ".\n\n";

  }


| LEMMA NAME DDOT term DOT {

    if (proof_mode()) then (

      printf "Error: actually already in proof mode\n\n";

    ) else (

      state.subgoals <- 0;
      let subgoal_name = (concat "" ("subgoal"::(string_of_int state.subgoals):: [])) in 
        if (typecheck2 state.ctxt $4 Type (listdef2def state.def) VarSet.empty) then (
          
          state.pt <- Some (([],subgoal_name, $4)::[], ($2, Var subgoal_name, $4));
          
          
          (*print_term ty; printf "\n\n";*)
          
            print_subgoals_ctxt();
            
        ) else (
          
          printf "Error: invalide type !\n\n";
          
        )
          
    );    
    
  }


| QED DOT {
    
    match state.pt with
      | None -> printf "Error: Not in proof mode\n\n";
      | Some ([], (n, te,ty)) ->
          let te2 = (simpl_red te VarSet.empty (listdef2def state.def)) in
            if (typecheck2 state.ctxt te2 ty (listdef2def state.def) VarSet.empty) then (
              
              printf "proof complete.\n\n";
              
              state.ctxt <- 
                (n, (readable_term_fp ty state.def))::state.ctxt;
              
              state.def <- 
                ((n, (readable_term_fp te2 state.def))::[])::state.def;
              
              state.pt <- None;
              state.subgoals <- 0;
              
            ) else (
              
              printf "Error: proof does not type!\n\n";
              print_term te2; printf "\n\n";
              (*print_term (simpl_red te VarSet.empty (listdef2def state.def)); printf "\n\n";*)
              (*print_term (readable_term_fp ty state.def); printf "\n\n";*)
              (*printf "%b\n\n" (typechecking te ty [] (listdef2def state.def));
                printf "%b\n\n" (alpha_typechecking te ty [] (listdef2def state.def));*)
              (*printf "%b\n\n" (typechecking ty Type [] (listdef2def state.def));*)
            )
      | _ -> 
          printf "proof not complete!\n\n";
          print_subgoals_ctxt();
  }

| SHOWGOAL DOT {

    match (state.pt) with
      | None -> printf "Error: Not in proof mode\n\n";
      | Some ((hd1,hd2,hd3)::tl,g) ->  (

          print_subgoals_ctxt();

        );
      | Some ([], g) -> printf "Error: Nomore subgoals\n\n";
    
  }

| PRINT term DOT {

    print_term (list_listrewrite_term_var_term $2 state.def); printf "\n\n";

  }



/***********************/
/* Tactics    commands */
/***********************/



| EQUIV term DOT {
    (
      match (state.pt) with
        | None -> printf "Error: Not in proof mode\n\n";
        | Some lg ->  (            
            match (equiv_tactic $2 lg (listdef2def state.def) state.ctxt) with
              | None -> printf "Error: No equivalence\n\n";
              | Some (lgh,lgg,n) ->
                  state.pt <- Some (lgh,lgg)
          )
    ); print_subgoals_ctxt();
  
  }




| INTRO DOT {
    (
      match (state.pt) with
        | None -> printf "Error: Not in proof mode\n\n";
        | Some ((hd1,hd2,Forall (v,t,hd3))::tl,(ln,te,ty)) ->  (
            match (intro_tactic v ((hd1,hd2,Forall (v,t,hd3))::tl,(ln,te,ty))) with
              | None -> printf "Error: No var to introduce\n\n";
              | Some (lgh,lgg,n) -> 
                  state.pt <- Some (lgh,lgg)
          );         
        | Some ([], g) -> printf "Error: Nomore subgoals\n\n";
        | _ -> printf "Error: No var to introduce\n\n";
    ); print_subgoals_ctxt();    
    
  }



| INTRO NAME DOT {
    (match (state.pt) with
       | None -> printf "Error: Not in proof mode\n\n";
       | Some lg -> (
           match (intro_tactic $2 lg) with
             | None -> printf "Error: No var to introduce\n\n";
             | Some (lgh,lgg,n) -> 
                 state.pt <- Some (lgh,lgg)
         )
    ); print_subgoals_ctxt();    

  }

| EXACT term DOT {

    match (state.pt) with
      | None -> printf "Error: Not in proof mode\n\n";
      | Some lg -> (
          match (exact_tactic $2 lg (listdef2def state.def) state.ctxt) with
            | None -> printf "Error: Invalide term.\n\n";
            | Some (lgh,lgg,n) ->
                state.pt <- Some (lgh,lgg)
        ); print_subgoals_ctxt();            
          
  }

| APPLY term DOT {

    (match (state.pt) with
       | None -> printf "Error: Not in proof mode\n\n";
       | Some lg -> (
           match (apply_tactic $2 lg (listdef2def state.def) state.ctxt) with
             | None -> printf "Error: Cannot apply\n\n";
             | Some (lgh,lgg,n) ->
                 state.pt <- Some (lgh,lgg)
         )
    ); print_subgoals_ctxt();
  }


| GENERALIZE term AS NAME DOT {
    
    (match (state.pt) with
       | None -> printf "Error: Not in proof mode\n\n";
       | Some lg -> 
           match (generalize_tactic $2 $4 lg (listdef2def state.def) state.ctxt) with
             | None -> printf "Error: Cannot generalize \n\n";
             | Some (lgh,lgg,n) ->
                 state.pt <- Some (lgh,lgg)
    ); print_subgoals_ctxt();                      

  }

| SIMPL DOT {
    (
      match state.pt with
        | None -> printf "Error: Not in proof mode\n\n";
        | Some lg ->
            match (simpl_tactic lg (listdef2def state.def)) with
              | None -> printf "Error: Nomore subgoals\n\n";
              | Some (lgh,lgg,n) ->
                  state.pt <- Some (lgh,lgg)
    ); print_subgoals_ctxt();
  }

| SIMPL IN NAME DOT {

    (match state.pt with
      | None -> printf "Error: Not in proof mode\n\n";
      | Some lg ->
          match (simpl_in_tactic $3 lg (listdef2def state.def)) with
            | None -> printf "Error: Cannot simplify\n\n";
            | Some (lgh,lgg,n) ->
                state.pt <- Some (lgh,lgg)
    ); print_subgoals_ctxt();
       
  }

| FACTORIZE term DOT {
    (
      match state.pt with
        | None -> printf "Error: Not in proof mode\n\n";
        | Some lg ->
            match (factorize_tactic $2 lg (listdef2def state.def) state.ctxt) with
              | None -> printf "Error: Cannot factorize\n\n";
              | Some (lgh,lgg,n) ->
                  state.pt <- Some (lgh,lgg)
    ); print_subgoals_ctxt();
      
  }

| DESTRUCT term DOT {
    (
      match state.pt with
        | None -> printf "Error: Not in proof mode\n\n";
        | Some lg ->
            match (destruct_tactic lg $2 (listdef2def state.def) state.ctxt) with
              | None -> printf "Error: Cannot destruct\n\n";
              | Some (lgh,lgg,n) ->
                  state.pt <- Some (lgh,lgg)
                    
    ); print_subgoals_ctxt();
      
  }

| UNFOLD NAME DOT {
    (
      match state.pt with
        | None -> printf "Error: Not in proof mode\n\n";
        | Some lg ->
            match (unfold_tactic lg $2 (listdef2def state.def)) with
              | None -> printf "Error: Cannot destruct\n\n";
              | Some (lgh,lgg,n) ->
                  state.pt <- Some (lgh,lgg)
                    
    ); print_subgoals_ctxt();
    

  }

| COMPUTE DOT {

    (match state.pt with
      | None -> printf "Error: Not in proof mode\n\n";
      | Some lg ->
          match (compute_tactic lg (listdef2def state.def)) with
            | None -> printf "Error: Cannot compute\n\n";
            | Some (lgh,lgg,n) ->
                state.pt <- Some (lgh,lgg)
    ); print_subgoals_ctxt();
    
  }

/***********************/
/* Debugs     commands */
/***********************/

/*

| LBRACK term RBRACK DOT {

    match (state.pt) with
      | None -> printf "Error: Not in proof mode\n\n";
      | Some ((hd1,hd2,hd3)::tl,(ln,te,ty)) ->  (
          
          (match (left_unification $2 hd3 (list_proj1_set hd1)) with
            | None -> printf "Error: No unification"
            | Some s -> (print_context (readable_context_fp (comp_subs s) state.def));
          );
          printf "\n\n";
            
          
        );
      | Some ([], g) -> printf "Error: Nomore subgoals\n\n";

  }

| LBRACK term RBRACK LPAREN definition RPAREN DOT {

    (print_term (readable_term_fp (apply_subst_term $2 $5 VarSet.empty) state.def)); printf "\n\n";

  }

| LBRACK definition RBRACK LPAREN definition RPAREN DOT {

    (print_context (readable_context_fp (apply_subst_context $2 $5 VarSet.empty) state.def)); printf "\n\n";

  }

*/

| QUIT {exit 0;}

;

  definition:
| LBRACK RBRACK {[]}
| LBRACK definition RBRACK { $2 }
| NAME EQ term SCOMMA { ($1, $3)::[] }
| LPAREN definition RPAREN { $2 }
| definition definition {$2 @ $1}
;


 term:
| terma {$1}
| term terma {App ($1,$2)}
;

  terma:
| NAME { 
    Var $1
  }
| LAMBDA LPAREN NAME DDOT term RPAREN DOT term {
    Lambda ($3, $5, $8)
  }
| FORALL LPAREN NAME DDOT term RPAREN DOT term {
    Forall ($3, $5, $8)
  }
| MATCH term WITH listterm {
    Match ($2, $4)
  }
| MATCH term RETURN term WITH listterm {
    MatchT ($2, $6, $4)
  }
| MATCH term WITH destructors {
    Match ($2, $4)
  }
| MATCH term RETURN term WITH destructors {
    MatchT ($2, $6, $4)
  }
| CONS NUM OF term {
    Cons ($2, $4)
  }
| TYPE { Type }
| LPAREN term RPAREN { $2 }
;

  listterm:
| term { $1 :: [] }
| listterm COMMA listterm { $1 @ $3 }
| LPAREN listterm RPAREN { $2 }
| LPAREN RPAREN { [] }
;

  context:
| LPAREN NAME DDOT term RPAREN { ($2, $4)::[] }
| context COMMA context { $1 @ $3 }
| LPAREN context RPAREN { $2 }
| LPAREN RPAREN { [] }
;

  args:

| LPAREN NAME DDOT term RPAREN { ($2,$4) }

;

  listargs:

| args { $1::[] }
| args listargs { $1 :: $2}

;

  constructors:
| constructors constructors { $1 @ $2 }
| PAR NAME DDOT term { ($2, $4)::[] }
;

 destructors:
| destructors destructors {$1 @ $2 }
| PAR term { $2::[] }
;

 listnum:
| NUM { $1::[] }
| listnum COMMA listnum {$1 @ $3}
;


