%{
(*
 This file is part of Poussin.

 Poussin is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Poussin is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Poussin.  If not, see <http://www.gnu.org/licenses/>.

 Copyright (C) 2008 Nicolas Marti
*)

  open Printf;;
  open Lexing;;
  open List;;
  open String;;
  
  open Def;;
  open Pprinter;;
  open Listutil;;
  open Freshness;;
  open Term;;
  open Alpharel;;
  open Betared;;
  open Termequiv;;
  open Typechecking
  open Unfold;;
  open Varset;;
  open Kernel;;

%}

/*  */
%token NEWLINE EOF

/* The indentifiers variables */
%token <string> NAME

/* Integer (for indexes of contructors) */
%token <int> NUM
  
/* quantifier */
%token LAMBDA IND FORALL PHI
  
/* keywords */
%token MATCH RETURN WITH CONS OF TYPE
  
/* parenthese */
%token LPAREN RPAREN LBRACK RBRACK

/* ponctuation */
%token SCOMMA COMMA DOT DDOT TURNSTIL EQ
  
/* kernel functions */
%token TERMEQ INFER

%start input
%type <unit> input

%% /* Grammar rules and actions follow */

  input:    

| request input {}

| EOF {}
  ;               


 term:
| terma {$1}
| term terma {App ($1,$2)}
;

  terma:
| NAME { 
    Var $1
  }
| LAMBDA LPAREN NAME DDOT term RPAREN DOT term {
    Lambda ($3, $5, $8)
  }
| FORALL LPAREN NAME DDOT term RPAREN DOT term {
    Forall ($3, $5, $8)
  }
| PHI LPAREN NAME DDOT term COMMA NUM RPAREN DOT term {
    Phi ($3, $7, $5, $10)
  }
| IND LPAREN NAME DDOT term RPAREN DOT listterm {
    Ind ($3, $5, $8)
  }
| MATCH term WITH listterm {
    Match ($2, $4)
  }
| MATCH term RETURN term WITH listterm {
    MatchT ($2, $6, $4)
  }
| CONS NUM OF term {
    Cons ($2, $4)
  }
| TYPE { Type }
| LPAREN term RPAREN { $2 }
;



  listterm:
| term { $1 :: [] }
| listterm COMMA listterm { $1 @ $3 }
| LPAREN listterm RPAREN { $2 }
| LPAREN RPAREN { [] }
;

  context:
| LPAREN NAME DDOT term RPAREN { ($2, $4)::[] }
| context COMMA context { $1 @ $3 }
| LPAREN context RPAREN { $2 }
| LPAREN RPAREN { [] }
;


  request:

| definition context TURNSTIL term DDOT term SCOMMA { 
    
    printf "type checking !!!\n";
    printf "--------------------\n";
    let def = (rev $1) in
    let ctxt = (listrewrite_context_var_term $2 def) in
    let te = (listrewrite_term_var_term $4 def) in
    let ty = (listrewrite_term_var_term $6 def) in
      print_context $2; printf "|-\n\n"; 
      print_term $4; printf "\n\n : "; 
      print_term $6; printf "\n\n";
      printf "--------------------\n";
      printf "result:= %b\n\n" (typecheck ctxt te ty def VarSet.empty);
    
  }

| definition context TURNSTIL term INFER  SCOMMA { 
    
    printf "type guessing !!!\n";
    printf "--------------------\n";
    let def = (rev $1) in
    let ctxt = (listrewrite_context_var_term $2 def) in
    let te = (listrewrite_term_var_term $4 def) in
      print_context $2; printf "|-\n\n"; print_term $4; printf "\n\n : ???\n\n";
      printf "--------------------\n";
      match (typeinfer ctxt te def VarSet.empty) with
        | None -> printf "Cannot infer!\n"
        | Some ty -> (
            printf "result:= "; 
            print_term (listfold_term_term_var ty $1);
            printf "\n\n"; 
            printf " (context) |- te : ty ? %b \n" (typecheck ctxt te ty def VarSet.empty);
            printf "--------------------\n\n";
            
          )
  }

| definition term  SCOMMA { 
    
    printf "term execution !!!\n";
    printf "--------------------\n";
    let def = (rev $1) in
    let te = (listrewrite_term_var_term $2 def) in
      print_term $2; printf ": ???\n\n";
      printf "--------------------\n";
      print_term (listfold_term_term_var (exec te def VarSet.empty) $1); printf "\n\n";


  }


| definition term TERMEQ SCOMMA { 
    
    printf "fold !!!\n";
    printf "--------------------\n";
    let def = rev $1 in 
    let te = (listrewrite_term_var_term $2 def) in
      print_term $2; printf " == \n\n"; 
      print_term te; printf " == \n\n";
      print_term (listfold_term_term_var (simpl te def VarSet.empty) $1); printf "\n\n";
      printf "--------------------\n";
  }

| definition context TURNSTIL term TERMEQ term  SCOMMA { 
    
    printf "type equiv alpha_rel !!!\n";
    printf "--------------------\n";
    printf "%b\n\n" (equiv $4 $6 $1 (context_to_setvar $2));


  }

;

  definition:
| LBRACK RBRACK {[]}
| LBRACK definition RBRACK { $2 }
| NAME EQ term SCOMMA { ($1, $3)::[] }
| LPAREN definition RPAREN { $2 }
| definition definition {$1 @ $2}
;


