(*
 This file is part of Poussin.

 Poussin is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Poussin is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Poussin.  If not, see <http://www.gnu.org/licenses/>.

 Copyright (C) 2008 Nicolas Marti
*)


open Printf;;
open List;;
open String;;
open Def;;

let rec print_term t =
  match t with
    | Var x -> printf "%s" x;
    | Lambda (x,t1,t2) -> printf "\\ (%s:" x; print_term t1; printf "). ("; print_term t2; printf ")";
    | Forall (x,t1,t2) -> printf "V (%s:" x; print_term t1; printf "). ("; print_term t2; printf ")";
    | Phi (x,n,t1,t2) -> printf "! (%s:" x; print_term t1; printf ", %d). (" n; print_term t2; printf ")";
    | Ind (x,t1,l2) -> printf "I (%s:" x; print_term t1; printf "). ("; print_listterm l2; printf ")";
    | App (t1, t2) -> (
        match t1 with
          | Var v1 -> print_term t1; 
          | App (t11, t12) -> print_term t1;
          | _ -> printf "("; print_term t1; printf ")";
        ); printf " "; (
        match t2 with
          | Var v2 -> print_term t2; 
          | _ -> printf "("; print_term t2; printf ")";
        );
    | Match (t1, l1) -> printf "match "; print_term t1; printf " with ("; print_listterm l1; printf ")";
    | MatchT (t1, l1, t2) -> printf "match "; print_term t1; printf " return "; print_term t2 ;printf " with ("; print_listterm l1; printf ")";
    | Cons (n,l) -> printf "cons %d of (" n; print_term l; printf ")";
    | Type -> printf "type"
and print_listterm l =
  match l with
    | [] -> printf "";
    | hd :: tl -> 
        print_term hd;
        match tl with
          | [] -> printf "";
          | _ -> printf ", "; print_listterm tl;;


let rec print_context c =
  match c with
    | [] -> printf "";
    | (hd1, hd2) :: tl ->
        printf "(%s: " hd1; 
        print_term hd2;
        printf ")"; 
        match tl with
          | [] -> printf "";
          | _ -> printf ", "; print_context tl;;


let rec print_list l =
  match l with
    | [] -> printf "";
    | hd :: tl -> 
        printf "%s" hd;
        match tl with
          | [] -> printf "";
          | _ -> printf ", "; print_list tl;;


let rec print_def l =
  match l with
    | [] -> printf "\n";
    | (hd1, hd2) :: tl ->
        printf "%s = " hd1; print_term hd2; printf ";\n";
        print_def tl;;


let rec print_subgoal_context c =
  match c with
    | [] -> printf "";
    | (hd1, hd2) :: tl ->
        printf "%s : " hd1; print_term hd2; printf "\n";
        print_subgoal_context tl;;

