(*
 This file is part of Poussin.

 Poussin is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Poussin is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Poussin.  If not, see <http://www.gnu.org/licenses/>.

 Copyright (C) 2008 Nicolas Marti
*)
  open Def;;
  open Pprinter;;
  open Listutil;;
  open Freshness;;
  open Term;;
  open Alpharel;;
  open Betared;;
  open Termequiv;;
  open Varset;;
  open Unfold;;

  (* rewriting of a subterm st by a term y in the term t (smooth) *)
  let rec rewrite_term_term_term t st y =
    if (term_eq t st) then y else (
      match t with
        | Lambda (s, t1, t2) ->
            if (st = Var s) then
              t else
              (Lambda (s, rewrite_term_term_term t1 st y, rewrite_term_term_term t2 st y))
        | Forall (s, t1, t2) ->
            if (st = Var s) then
              t else
              (Forall (s, rewrite_term_term_term t1 st y, rewrite_term_term_term t2 st y))
        | Phi (s, n, t1, t2) ->
            if (st = Var s) then
              t else
              (Phi (s, n, rewrite_term_term_term t1 st y, 
                   match st with
                     | Var x ->
                         if (VarSet.subset (VarSet.singleton x) (s ++ (context_to_setvar (forall_list_var_type t1))))
                         then t2 else
                           rewrite_term_term_term t2 st y
                     | _ -> rewrite_term_term_term t2 st y
              )
              )
        | Ind (s, t1, l) ->
            if (st = Var s) then
              t else
              (Ind (s, rewrite_term_term_term t1 st y, rewrite_listterm_term_term l st y))
        | App (t1, t2) ->
            App (rewrite_term_term_term t1 st y, rewrite_term_term_term t2 st y)
        | Match (t1, l) ->
            Match (rewrite_term_term_term t1 st y, rewrite_listterm_term_term l st y)
        | MatchT (t1, l, t2) ->
            MatchT (rewrite_term_term_term t1 st y, rewrite_listterm_term_term l st y, t2)
        | Cons (n, l) ->
            Cons (n, rewrite_term_term_term l st y)
        | _ -> t
    )
  and rewrite_listterm_term_term l x y =
    match l with
      | [] -> []
      | hd :: tl ->
          (rewrite_term_term_term hd x y) :: (rewrite_listterm_term_term tl x y);;


  let rec listrewrite_term_term_term t l =
    match l with
      | [] -> t
      | (hd1, hd2) :: tl ->
          listrewrite_term_term_term (rewrite_term_term_term t hd1 hd2) tl;;

  let rec rewrite_context_term_term c x y =
    match c with
      | [] -> []
      | (hd1, hd2) :: tl ->
          (hd1, rewrite_term_term_term hd2 x y) :: (rewrite_context_term_term tl x y);;
