(*
 This file is part of Poussin.

 Poussin is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Poussin is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Poussin.  If not, see <http://www.gnu.org/licenses/>.

 Copyright (C) 2008 Nicolas Marti
*)


open Printf;;
open Lexing;;
open List;;
open String;;

open Def;;
open Pprinter;;
open Listutil;;
open Freshness;;
open Term;;
open Alpharel;;
open Betared;;
open Termequiv;;
open Typechecking
open Unfold;;
open Proofbuild;;
open Varset;;
open Unification;;
open Rewriting;;
open Kernel;;

(*******************************)
(* Subroutines used by tactics *)
(*******************************)

let rec rewrite_subgoal_var_term sg x t =
  match sg with 
    | [] -> []
    | (ctxt, sgn, sgt) :: tl ->
        (ctxt, sgn, rewrite_term_var_term sgt x t) :: (rewrite_subgoal_var_term tl x t);;

let rec apply_modifie_state' t q s ctxt sv =
  match q with
    | [] -> ([],t)
    | (hd1, hd2) :: tl ->
        match (var_type_in_context hd1 s) with
          | None -> (
              let x = (fresh_name_list_name sv "subgoal0") in 
              let (c, te) = (apply_modifie_state' (App(t,Var x)) tl s ctxt (x ++ sv)) in
                ((ctxt,x,hd2):: (rewrite_subgoal_var_term c hd1 (Var x)) , te)
            )
          | Some te ->
              apply_modifie_state' (App(t,te)) tl s ctxt sv;;


let apply_modifie_state t q s hd1 hd2 hd3 tl ln te ty =
  let (nsg, pt) = (apply_modifie_state' t q s hd1 (term_set_var te)) in          
  let te2 = (rewrite_term_var_term te hd2 pt) in (
      (nsg @ tl, (ln, te2, ty), List.length nsg)
    );;
 
let rec simpl_ctxt c x def =
  match c with 
    | [] -> []
    | (hd1, hd2) :: tl ->
        if (hd1 = x) then (hd1,readable_term_fp (simpl (listrewrite_term_var_term hd2 def) def VarSet.empty) (def::[])) :: tl else
          (hd1, hd2) :: (simpl_ctxt tl x def);;


let rec destruct_build n iname itype ilc lc args bv ty ctxt dt def =
  match lc with
    | [] -> ([], [])
    | (hdc::tlc) -> (
        let hdc2 = (cf_term_listvar hdc bv) in
        let (con, fcv) = (forall_concl_list_var_type hdc2) in 
          let (i, iargs) = (app_first_term_args con) in 
            match (unification_list 
                      (listrewrite_listterm_var_term iargs def)
                      (listrewrite_listterm_var_term args def)
                      VarSet.empty) with
              | NoMgu -> 
                  destruct_build (n+1) iname itype ilc tlc args bv ty ctxt dt def
              | Mgu s ->
                  let hdc3 = (cf_term_listvar hdc (context_to_setvar (fcv @ ctxt))) in
                  let (_, lvt) = (forall_concl_list_var_type hdc3) in 
                  let s' = comp_subs (
                    
                    match (unification_list (list_name2list_var (list_proj1 fcv)) (list_name2list_var (list_proj1 lvt)) VarSet.empty) with
                      | Mgu s2 -> s2
                      | _ -> []
                  ) in
                  let s5 = comp_subs (apply_subst_context (match_comp_args s fcv lvt) s' VarSet.empty) in
                  let sf = s5 in
                    
                  let ctxt' = (rev(lvt) @ ctxt) in
                  let ctxt'' = (rewrite_context_term_term ctxt' dt (listargs_app (Cons (n,Ind(iname, itype, ilc))) (list_name2list_var (list_proj1 lvt)))) in
                  let ctxt''' = (apply_subst_context ctxt'' sf VarSet.empty) in
                    
                  let ty' = (rewrite_term_term_term ty dt (listargs_app (Cons (n,Ind(iname, itype, ilc))) (list_name2list_var (list_proj1 lvt)))) in
                  let ty'' = (apply_subs_term ty' sf VarSet.empty) in
                    
                  
                  let hdc' = (rewrite_term_term_term hdc3 dt (listargs_app (Cons (n,Ind(iname, itype, ilc))) (list_name2list_var (list_proj1 lvt)))) in
                  let hdc'' = (apply_subs_term hdc' sf VarSet.empty) in
                  let (_, lvt') = (forall_concl_list_var_type hdc'') in 
                  let sgn = (fresh_name_list_name bv "subgoal0") in 
                    
                    
                  let (l1, l2) = (destruct_build (n + 1) iname itype ilc tlc args (sgn ++ bv) ty ctxt dt def) in
                    ((ctxt''', sgn, ty'')::l1, (concl_list_var_type_lambda (Var sgn) lvt')::l2)
                      
              | DnkMgu ->
                  let hdc3 = (cf_term_listvar hdc (context_to_setvar (fcv @ ctxt))) in
                  let (_, lvt) = (forall_concl_list_var_type hdc3) in 
                  let s' = comp_subs (
                    
                    match (unification_list (list_name2list_var (list_proj1 fcv)) (list_name2list_var (list_proj1 lvt)) VarSet.empty) with
                      | Mgu s2 -> s2
                      | _ -> []
                  ) in
                  let sf = s' in
                    
                  let ctxt' = (rev(lvt) @ ctxt) in
                  let ctxt'' = (rewrite_context_term_term ctxt' dt (listargs_app (Cons (n,Ind(iname, itype, ilc))) (list_name2list_var (list_proj1 lvt)))) in
                  let ctxt''' = (apply_subst_context ctxt'' sf VarSet.empty) in
                    
                  let ty' = (rewrite_term_term_term ty dt (listargs_app (Cons (n,Ind(iname, itype, ilc))) (list_name2list_var (list_proj1 lvt)))) in
                  let ty'' = (apply_subs_term ty' sf VarSet.empty) in
                    
                  
                  let hdc' = (rewrite_term_term_term hdc3 dt (listargs_app (Cons (n,Ind(iname, itype, ilc))) (list_name2list_var (list_proj1 lvt)))) in
                  let hdc'' = (apply_subs_term hdc' sf VarSet.empty) in
                  let (_, lvt') = (forall_concl_list_var_type hdc'') in 
                  let sgn = (fresh_name_list_name bv "subgoal0") in 
                    
                    
                  let (l1, l2) = (destruct_build (n + 1) iname itype ilc tlc args (sgn ++ bv) ty ctxt dt def) in
                    ((ctxt''', sgn, ty'')::l1, (concl_list_var_type_lambda (Var sgn) lvt')::l2)

      );;


(**************************)
(* Tactics main functions *)
(**************************)


(* equivalence *)

let equiv_tactic t lg def ctxt =

  match lg with
    | ((hd1,hd2,hd3)::tl,g) -> (
        if (equiv t hd3 def (context_to_setvar (hd1 @ ctxt))) then (      
          Some ((hd1,hd2,t)::tl,g,1)            
        ) else (          
          None
        )
      )
    | _ -> None
;;           



(* intro *)

let intro_tactic x lg =
  match lg with
    | ((hd1,hd2,Forall (v,t,hd3))::tl,(ln,te,ty)) ->  (

        let x2 = (fresh_name_list_name (context_set_var (hd1 @ state.ctxt)) x) in (
            Some (((x2,t)::hd1,hd2,(rewrite_term_var_term hd3 v (Var x2)))::tl,
                  (ln,rewrite_term_var_term te hd2 (Lambda (x2,t,(Var hd2))),ty), 1)
          )
                  
      );
    | _ -> None
;;


(* exact *)
        
let exact_tactic t lg def ctxt =
  match lg with
    | ((hd1,hd2,hd3)::tl,(ln,te,ty)) ->  (
        
        if (typecheck2
              (hd1 @ ctxt)
              t
              hd3 
              def
              (context_to_setvar (hd1 @ ctxt))
           ) then (
          Some (rewrite_subgoal_var_term tl hd2 t,(ln,(rewrite_term_var_term te hd2 t),ty), 0)
        ) else None
      )
    | _ -> None
;;



(* apply *)
          
let apply_tactic t lg def ctxt =

    match lg with
      | ((hd1,hd2,hd3)::tl,(ln,te,ty)) ->  (
                    
          match (typeinfer2 
                   (hd1 @ ctxt)
                   t                    
                   def
                   VarSet.empty
                ) with
            | None -> None
            | Some ty2 -> 
                if (equiv 
                      ty2 
                      hd3 
                      def
                      (context_to_setvar (hd1 @ ctxt)) 
                   ) then (
                  
                  Some 
                    (rewrite_subgoal_var_term tl hd2 t,
                     (ln,
                      (
                        rewrite_term_var_term te hd2 t),ty
                     ),
                     0
                    )

                  
                ) else (
                  let ty2' = (cf_term_listvar (beta_red ty2 (context_to_setvar hd1) []) (context_to_setvar hd1)) in
                  let (c, q) = (forall_concl_list_var_type ty2') in
(*                    print_term (readable_term_fp ty2' state.def); printf "\n\n";
                    print_context (readable_context_fp q state.def); printf "\n\n";
                    print_term (readable_term_fp c state.def); printf "\n\n";
                    print_term (readable_term_fp hd3 state.def); printf "\n\n";*)
                    match (unification c hd3 (list_proj1_set hd1)) with
                      | Mgu s -> 
                          (*print_context (readable_context_fp (comp_subs s) state.def); printf "\n\n";*)
                          let q' = (apply_subst_context q (comp_subs s) VarSet.empty) in 
(*                            print_context (readable_context_fp q' state.def); printf "\n\n";
                            print_term (readable_term_fp t state.def); printf "\n\n";*)
                            Some (apply_modifie_state t q' (comp_subs s) hd1 hd2 hd3 tl ln te ty)
                      | NoMgu -> None                              
                      | DnkMgu -> printf "DnkMgu\n"; None                              
                )
        )
      | _ -> None;;



(* generalize *)      

let generalize_tactic t x lg def ctxt =
  match lg with
    | ((hd1,hd2,hd3)::tl,(ln,te,ty)) ->  (
        
        (
          match (typeinfer2 
                   (hd1 @ ctxt)
                   t
                   def
                   VarSet.empty)
          with
            | None -> None
            | Some ty2 -> 
                let hd3' = (Forall (x,ty2, rewrite_term_term_term hd3 t (Var x))) in (
                    
                    Some ((hd1,hd2,hd3')::tl,
                          (ln,rewrite_term_var_term te hd2 (App(Var hd2, t)),ty),
                          1
                         )
                  )                    
        )
          
      )
    | _ -> None

(* simpl *)
      
let simpl_tactic lg def =
  match lg with
    | ((hd1,hd2,hd3)::tl,(ln,te,ty)) ->  (
        (*Some ((hd1,hd2, readable_term_fp (simpl (listrewrite_term_var_term hd3 def) def VarSet.empty) (def::[]))::tl,(ln,te,ty), 1)*)
        Some ((hd1,hd2, readable_term_fp (simpl (listrewrite_term_var_term hd3 def) def VarSet.empty) (def::[]))::tl,(ln,te,ty), 1)
      )
    | _ -> None
;; 

(* simpl in *)        

let simpl_in_tactic x lg def =
    match lg with
      | ((hd1,hd2,hd3)::tl,(ln,te,ty)) ->  (
          Some ((simpl_ctxt hd1 x def,hd2,hd3)::tl,(ln,te,ty), 1)
        )
      | _ -> None
;;
            
(* factorize *)
      
let factorize_tactic t lg def ctxt =
  match lg with
    | ((hd1,hd2,hd3)::tl,(ln,te,ty)) ->  (
          
        let x = (fresh_name_list_name ((context_set_var (hd1 @ ctxt)) +++ (term_set_var hd3)) "f") in (
            
            match (
              typeinfer2
                (hd1 @ ctxt)
                t
                def
                (context_to_setvar (hd1 @ctxt))
            ) with
              | None -> None
              | Some ty2 ->                   
                  Some ((hd1,hd2,App(Lambda (x,ty2,rewrite_term_term_term hd3 t (Var x)), t))::tl,(ln,te,ty), 1)
          )
      )
    | _ -> None 
;;          

(* destruct *)

let destruct_tactic lg dt def ctxt =
  match lg with
    | ((hd1,hd2,hd3)::tl,(ln,te,ty)) ->  (
        match (typeinfer2 (hd1 @ ctxt) dt def (context_to_setvar (hd1 @ ctxt))) with
          | None -> None
          | Some t1' -> (
              let (t1'', args) = (app_first_term_args (listrewrite_term_var_term t1' (rev def))) in
                match (listrewrite_term_var_term t1'' def) with
                  | Ind (x,t11,t12) -> (
                      
                      (*
                        
                        lsg = list of the new subgoals (ctxt, name, type)
                        lmd = list of match destructor rules
                        
                        match dt return type_sg with lmd
                  
                      *)
                      
                      let (lsg, lmd) = (destruct_build 
                                          0
                                          x 
                                          t11 
                                          t12 
                                          t12
                                          args 
                                          (context_to_setvar hd1 +++ term_set_var te)
                                          hd3
                                          hd1
                                          (listrewrite_term_var_term dt def)
                                          def
                                       ) in (
                          Some (
                            lsg @ tl,
                            (ln, rewrite_term_var_term te hd2 (MatchT (dt,lmd,hd3)), ty),
                            List.length lsg
                          )
                        )
                    )
                  | _ -> None
            )
      )
    | _ -> None
;;

(* unfold *)
let unfold_tactic lg n def =
  match lg with
    | ((hd1,hd2,hd3)::tl,(ln,te,ty)) ->  (

        Some ((hd1,hd2,unfold_fct hd3 n def)::tl,(ln,te,ty), 1)

      )
    | _ -> None
;;

(* compute *)
      
let compute_tactic lg def =
  match lg with
    | ((hd1,hd2,hd3)::tl,(ln,te,ty)) ->  (
        Some (((hd1,hd2,readable_term_fp (exec (listrewrite_term_var_term hd3 def) def VarSet.empty) (def::[])) ::tl),(ln,te,ty), 1)
      )
    | _ -> None
;; 
            
