(*
 This file is part of Poussin.

 Poussin is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Poussin is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Poussin.  If not, see <http://www.gnu.org/licenses/>.

 Copyright (C) 2008 Nicolas Marti
*)

open List;;
open String;;
open Printf;;

open Varset;;
open Def;;
open Pprinter;;
open Listutil;;
open Freshness;;
open Term;;
open Alpharel;;
open Betared;;
open Termequiv;;
open Typechecking;;
open Unification;;
open Unfold;;
open Arith;;

open Rewriting;;

let rec rec_args_list_term args indexes =
  match indexes with
    | [] -> Some []
    | hd :: tl -> 
        match (rec_args_list_term args tl) with
          | None -> None
          | Some tl' ->
              match (nth args hd) with
                | None -> None
                | Some x -> Some (x::tl');;

let rec term_size t =
    match t with
    | Var v -> 
        Some (Nvar v)
    | App (t1, t2) -> (
        match (term_size t1, term_size t2) with
          | (Some s1, Some s2) -> Some (Nplus (s1, s2))
          | _ -> None
      )
    | Cons (n, t1) ->
        Some (Nval 1)
    | Ind (x, t1, t2) -> Some (Nval 1)
    | _ -> None;;


let rec recargs_form args vars =
  match (args, vars) with
    | ([],[]) -> Some True
    | (hd1::tl1, hd2::tl2) -> (
        match (recargs_form tl1 tl2) with
          | None -> None
          | Some tl' ->
              match (term_size hd1, term_size hd2) with
                | (Some s1, Some s2) -> Some (And (Gt(s2, s1), tl'))
                | _ -> None
      )
    | _ -> None;;

let rec sum_size_list_var l =
  match l with
    | [] -> Nval 0
    | hd :: tl ->
        Nplus (Nvar hd, sum_size_list_var tl);;

let rec sum_size_list_term l =
  match l with
    | [] -> Some (Nval 0)
    | hd :: tl ->
        match (sum_size_list_term tl, term_size hd) with
          | Some tl', Some hd' ->
              Some (Nplus (hd', tl'))
          | _ -> None;;
        
let rec and_gt0_list_var l =
  match l with
    | [] -> True
    | hd :: tl ->
        And (Gt (Nvar hd, Nval 0), and_gt0_list_var tl);;

(* version for dependant type *)
let rec term_build_term_form t ctxt args indexes recname def =  
  
  if (not (VarSet.subset (VarSet.singleton recname) (term_fv t (context_to_setvar ctxt)))) then True else (    
    (*print_context ctxt; printf "\n"; print_term t; printf "\n\n"; *)
    match t with
      | Lambda (x, tx, t') ->
          term_build_term_form t' ((x,tx) :: ctxt) args indexes recname def
      | Forall (x, tx, t') ->
          And (
            term_build_term_form tx ctxt args indexes recname def,
            term_build_term_form t' ctxt args indexes recname def
          )
      | MatchT (x, ld, t') ->
          term_build_term_form (Match (x,ld)) ctxt args indexes recname def
      | App (t1, t2) ->
          let (t_fct, t_args) = (app_first_term_args t) in 
            if (t_fct = Var recname) then (
              (*
                print_term t_fct;
                printf "\n\n";             
              *)
              match (rec_args_list_term t_args indexes, rec_args_list_term args indexes) with
                | (Some vars1, Some vars2) -> (
                    
                    match (sum_size_list_term vars2, sum_size_list_term vars1) with
                      | (Some s1, Some s2) ->
                          Gt (s1, s2)
                      | _ -> 
                          printf "False 1\n\n"; 
                          print_listterm vars1;
                          printf "\n";
                          print_listterm vars2;
                          printf "\n";
                          False                          

                  )
                | _ -> printf "False 2\n\n"; False
            ) else (
              (*
                print_term t_fct;
                printf " =/= ";
                print_term (Var recname);
                printf "\n\n";
                
                term_build_listterm_form t_args ctxt args indexes recname
              *)

              And (
                term_build_term_form t1 ctxt args indexes recname def, 
                term_build_term_form t2 ctxt args indexes recname def
              )
            )   
      | Match (x, ld) -> (
          match (infer x ctxt [] simpl_red) with
            | None -> printf "False 3\n\n"; False
            | Some t1' -> 
              let (t1'_fct, t1'_args) = (app_first_term_args (simpl_red t1' VarSet.empty def)) in 
                (
                  match t1'_fct with 
                    | (Ind(x2,t11,t12)) -> (
                        term_build_term_form_match x ld t12 t1'_args args indexes ctxt recname 0 (Ind(x2,t11,t12)) def
                      )
                    | _ -> (
                        match (listrewrite_term_var_term t1'_fct def) with 
                          | (Ind(x2,t11,t12)) -> (
                              term_build_term_form_match x ld t12 t1'_args args indexes ctxt recname 0 (Ind(x2,t11,t12)) def
                            )
                          | t1'' -> 
                              printf "False 4\n"; 
                              print_term x; printf "\n\n";
                              print_term t1'; printf "\n\n";
                              print_term t1''; printf "\n\n";
                              False
                      )
                )
        )      
      | _ -> False
  )

 and term_build_term_form_match x ld lc cargs args indexes ctxt recname n ind def =
  match (lc, ld) with
    | ([], []) -> True
    | (hdc::tlc, hdd::tld) -> (
        let hdc2 = (cf_term_listvar hdc (context_to_setvar ctxt)) in
        let (con, fcv) = (forall_concl_list_var_type hdc2) in 
        let (i, iargs) = (app_first_term_args con) in 
        match (unification_list 
                  (listrewrite_listterm_var_term iargs def)
                  (listrewrite_listterm_var_term cargs def)
                  VarSet.empty) with
          | NoMgu -> 
              term_build_term_form_match x ld tlc cargs args indexes ctxt recname (n+1) ind def
          | Mgu s ->
              let hdd2 = (cf_term_listvar hdd (context_to_setvar (fcv @ ctxt))) in
              let (_, lvt) = (lambda_concl_list_var_type hdd2 (List.length fcv)) in 
              let s' = comp_subs (
                
                match (unification_list (list_name2list_var (list_proj1 fcv)) (list_name2list_var (list_proj1 lvt)) VarSet.empty) with
                  | Mgu s2 -> s2
                  | _ -> []
              ) in
              let s5 = comp_subs (apply_subst_context (match_comp_args s fcv lvt) s' VarSet.empty) in
                
              let ctxt' = (rev(lvt) @ ctxt) in
              let hdd' = (rewrite_term_term_term hdd2 x (listargs_app (Cons (n,ind)) (list_name2list_var (list_proj1 lvt)))) in
                
              let (hdd'', _) = (lambda_concl_list_var_type hdd' (List.length fcv)) in 
              let ctxt'' = (rewrite_context_term_term ctxt' x (listargs_app (Cons (n,ind)) (list_name2list_var (list_proj1 lvt)))) in
              let sf = s5 in
              let ctxt''' = (apply_subst_context ctxt'' sf VarSet.empty) in
              let hdd''' = (apply_subs_term hdd'' sf VarSet.empty) in (
                (*print_context ctxt'; printf "\n"; print_term t'; printf "\n\n";*)
                match (term_size x) with
                  | None -> printf "False 5\n\n"; False
                  | Some sx ->
                      And (
                        Impl (
                          Eq(sx, Nplus (Nval 1, sum_size_list_var (list_proj1 lvt)))
                          ,
                          term_build_term_form hdd''' ctxt''' args indexes recname def
                        ),
                        term_build_term_form_match x tld tlc cargs args indexes ctxt recname (n+1) ind def
                      )
              )
          | DnkMgu ->
              let hdd2 = (cf_term_listvar hdd (context_to_setvar (fcv @ ctxt))) in
              let (_, lvt) = (lambda_concl_list_var_type hdd2 (List.length fcv)) in 
              let s' = comp_subs (
                
                match (unification_list (list_name2list_var (list_proj1 fcv)) (list_name2list_var (list_proj1 lvt)) VarSet.empty) with
                  | Mgu s2 -> s2
                  | _ -> []
              ) in
              let ctxt' = (rev(lvt) @ ctxt) in
              let hdd' = (rewrite_term_term_term hdd2 x (listargs_app (Cons (n,ind)) (list_name2list_var (list_proj1 lvt)))) in
                
              let (hdd'', _) = (lambda_concl_list_var_type hdd' (List.length fcv)) in 
              let ctxt'' = (rewrite_context_term_term ctxt' x (listargs_app (Cons (n,ind)) (list_name2list_var (list_proj1 lvt)))) in
              let sf = s' in
              let ctxt''' = (apply_subst_context ctxt'' sf VarSet.empty) in
              let hdd''' = (apply_subs_term hdd'' sf VarSet.empty) in (
                (*print_context ctxt'; printf "\n"; print_term t'; printf "\n\n";*)
                match (term_size x) with
                  | None -> printf "False 5\n\n"; False
                  | Some sx ->
                      And (
                        Impl (
                          Eq(sx, Nplus (Nval 1, sum_size_list_var (list_proj1 lvt)))
                          ,
                          term_build_term_form hdd''' ctxt''' args indexes recname def
                        ),
                        term_build_term_form_match x tld tlc cargs args indexes ctxt recname (n+1) ind def
                      )
              )
      )
    | _ -> printf "False 6\n\n"; False
;;
        

        

