(*
  This file is part of Poussin.

  Poussin is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Poussin is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Poussin.  If not, see <http://www.gnu.org/licenses/>.

  Copyright (C) 2008 Nicolas Marti
*)

open List;;
open String;;
open Printf;;

open Def;;
open Pprinter;;
open Listutil;;
open Freshness;;
open Term;;
open Alpharel;;
open Betared;;
open Termequiv;;
open Unfold;;
open Varset;;
open Rewriting;;
open Unification;;

let valid_destructor hdc x ind  args ctxt def =
  let hdc2 = (cf_term_listvar (subst_var_term hdc x ind (context_to_setvar ctxt)) (context_to_setvar ctxt)) in
  let (con, fcv) = (forall_concl_list_var_type hdc2) in 
  let (i, iargs) = (app_first_term_args con) in 
    match (unification_list 
              (listrewrite_listterm_var_term iargs def)
              (listrewrite_listterm_var_term args def) 
              VarSet.empty) with
      | NoMgu -> false
      | _ -> true
;;
        
(* a factorization for destruction of match destructor *)
let destruct_deptype n hdc hdd t x args ind ty ctxt def =
  let hdc2 = (cf_term_listvar (subst_var_term hdc x ind (context_to_setvar ctxt)) (context_to_setvar ctxt)) in
  let (con, fcv) = (forall_concl_list_var_type hdc2) in 
  let (i, iargs) = (app_first_term_args con) in 
    match (unification_list 
              (listrewrite_listterm_var_term iargs def)
              (listrewrite_listterm_var_term args def) 
              VarSet.empty) with
      | NoMgu -> None

      | DnkMgu ->
          (
            printf "DnkMgu\n";
            print_listterm (listrewrite_listterm_var_term iargs def); printf "\n";
            print_listterm (listrewrite_listterm_var_term args def); printf "\n\n";

            let hdd2 = (cf_term_listvar hdd (context_to_setvar (fcv @ ctxt))) in
            let (_, lvt) = (lambda_concl_list_var_type hdd2 (List.length fcv)) in 
            let s' = comp_subs (
              
              match (unification_list (list_name2list_var (list_proj1 fcv)) (list_name2list_var (list_proj1 lvt)) VarSet.empty) with
                | Mgu s2 -> s2
                | _ -> []
            ) in

            let ctxt' = (rev(lvt) @ ctxt) in
            let hdd' = (rewrite_term_term_term hdd2 t (listargs_app (Cons (n,ind)) (list_name2list_var (list_proj1 lvt)))) in
            let (hdd'', _) = (lambda_concl_list_var_type hdd' (List.length fcv)) in 
            let ctxt'' = (rewrite_context_term_term ctxt' t (listargs_app (Cons (n,ind)) (list_name2list_var (list_proj1 lvt)))) in
            let ty' = (rewrite_term_term_term ty t (listargs_app (Cons (n,ind)) (list_name2list_var (list_proj1 lvt)))) in
            let sf = s' in
            let ctxt''' = (apply_subst_context ctxt'' sf VarSet.empty) in
            let hdd''' = (apply_subs_term hdd'' sf VarSet.empty) in                        
            let ty'' = (apply_subs_term ty' sf VarSet.empty) in              
            let lvt' = (
              (rev (apply_subst_context (rev lvt) sf VarSet.empty))
            ) in 
            let fcv' = (
              (rev (apply_subst_context (rev fcv) sf VarSet.empty))
            ) in
              Some (
                (
                  (simpl_red hdd''' (context_to_setvar ctxt''') def), 
                  (simpl_red ty'' (context_to_setvar ctxt''') def),
                  ctxt'''
                ),
                (
                  (concl_list_var_type_lambda Type lvt'),
                  (concl_list_var_type_forall Type fcv'),
                  (apply_subst_context ctxt sf VarSet.empty)
                )
              )                          
          )                
      | Mgu s -> 
          (
            let hdd2 = (cf_term_listvar hdd (context_to_setvar (fcv @ ctxt))) in
            let (_, lvt) = (lambda_concl_list_var_type hdd2 (List.length fcv)) in 
            let s' = comp_subs (
              match (unification_list (list_name2list_var (list_proj1 fcv)) (list_name2list_var (list_proj1 lvt)) VarSet.empty) with
                | Mgu s2 -> s2
                | _ -> []
            ) in
            let s5 = comp_subs (apply_subst_context (match_comp_args s fcv lvt) s' VarSet.empty) in
            let ctxt' = (rev(lvt) @ ctxt) in
            let hdd' = (rewrite_term_term_term hdd2 t (listargs_app (Cons (n,ind)) (list_name2list_var (list_proj1 lvt)))) in
            let (hdd'', _) = (lambda_concl_list_var_type hdd' (List.length fcv)) in 
            let ctxt'' = (rewrite_context_term_term ctxt' t (listargs_app (Cons (n,ind)) (list_name2list_var (list_proj1 lvt)))) in
            let ty' = (rewrite_term_term_term ty t (listargs_app (Cons (n,ind)) (list_name2list_var (list_proj1 lvt)))) in
            let sf = s5 in
            let ctxt''' = (apply_subst_context ctxt'' sf VarSet.empty) in
            let hdd''' = (apply_subs_term hdd'' sf VarSet.empty) in                        
            let ty'' = (apply_subs_term ty' sf VarSet.empty) in
              
            let lvt' = (
              (rev (apply_subst_context (rev lvt) sf VarSet.empty))
            ) in 
            let fcv' = (
              (rev (apply_subst_context (rev fcv) s VarSet.empty))
            ) in
              Some (
                (
                  (simpl_red hdd''' (context_to_setvar ctxt''') def), 
                  (simpl_red ty'' (context_to_setvar ctxt''') def),
                  ctxt'''
                ),
                (
                  (concl_list_var_type_lambda Type lvt'),
                  (concl_list_var_type_forall Type fcv'),
                  (apply_subst_context ctxt sf VarSet.empty)
                )
              )                                    
          ) 
;;

(* type inference with a reduction exec *)

let rec infer t ctxt def exec =
  (*printf "("; print_context (listfold_context_term_var_fp ctxt def); printf ") |-\n\t"; print_term (listfold_term_term_var_fp t def); printf " : ???\n\n";*)
  match t with
    | (Var x) -> (
        match (var_type_in_context x ctxt) with
          | None -> (printf "infer(var) : %s\n\n" x; 
                     printf "("; print_context (listfold_context_term_var_fp ctxt def); printf ") |-\n\t"; print_term (listfold_term_term_var_fp t def); printf " : ???\n\n";
                     None)
          | Some i -> Some (exec i (context_to_setvar ctxt) def)
      )

    | Type -> Some Type

    | (Lambda (x,t1,t2)) -> (
        match (infer t2 ((x,t1)::ctxt) def exec) with
          | None -> None
          | Some ty -> Some (Forall (x,t1,ty))
      )

    | (Forall (x,t1,t2)) -> Some Type

    | (Phi (x, n, t1,t2)) -> Some t1

    | (Ind (x,t1,t2)) -> Some t1

    | (Cons (n, t1)) -> (
        
        match (exec t1 (context_to_setvar ctxt) def) with
          | (Ind (x,t1,t2)) -> (
              
              match (nth t2 n) with
                | None -> printf "infer(cons 1)\n\n"; None
                | Some t3 -> Some (subst_var_term t3 x (Ind (x, t1, t2)) (context_to_setvar ctxt))
            )

          | t1' -> (
              match (listrewrite_term_var_term t1' def) with
                | (Ind (x,t1,t2)) -> (
                    
                    match (nth t2 n) with
                      | None -> printf "infer(cons 2)\n\n"; None
                      | Some t3 -> Some (subst_var_term t3 x (Ind (x, t1, t2)) (context_to_setvar ctxt))
                  )
                    
                | t1'' -> (
                    printf "infer(cons 3)\n"; 
                    print_term t1''; printf "\n\n";
                    None
                  )

            )
      )

    | (App (t1, t2)) -> (
        match (infer t1 ctxt def exec, infer t2 ctxt def exec) with
          | (Some (Forall (x, t11, t12)), Some t2') ->
              
              if (term_equiv t11 t2' (context_to_setvar ctxt) def) then (
                
                Some (subst_var_term t12 x t2 (context_to_setvar ctxt)) 
                  
              )
              else (

                printf "infer(app 1)\n"; 
                printf "("; print_context (listfold_context_term_var_fp ctxt def); printf ") |-\n\t"; print_term (listfold_term_term_var_fp t def); printf " : ???\n\n";                                     
                printf "\t t11 = "; print_term (listfold_term_term_var_fp (exec t11 (context_to_setvar ctxt) def) def); printf "\n\n";
                printf "\t t12 = "; print_term (listfold_term_term_var_fp (exec t12 (context_to_setvar ctxt) def) def); printf "\n\n";
                printf "\t t2' = "; print_term (listfold_term_term_var_fp (exec t2' (context_to_setvar ctxt) def) def); printf "\n\n\n";

                (*printf "\t t11 = "; print_term (exec t11 (context_to_setvar ctxt)); printf "\n\n";
                  printf "\t t2' = "; print_term (exec t2' (context_to_setvar ctxt)); printf "\n";*)
                
                
                None
              )
          | (None, _) -> (None)
          | (_, None) -> (None)
          | (Some t1', Some t2') -> (
              match (simpl_red (listrewrite_term_var_term t1' def) VarSet.empty def) with
                | (Forall (x, t11, t12)) ->
                    
                    if (term_equiv t11 t2' (context_to_setvar ctxt) def) then (
                      (*
                        print_term (subst_var_term t12 x t2 (context_to_setvar ctxt));
                        printf "\n\n";
                      *)
                      Some (subst_var_term t12 x t2 (context_to_setvar ctxt)) 
                        
                    ) else
                      None
                | _ ->
                    printf "infer(app 2)\n"; 
                    printf "("; print_context (listfold_context_term_var_fp ctxt def); printf ") |-\n\t"; print_term (listfold_term_term_var_fp t def); printf " : ???\n\n";                                     
                    
                    printf "App ( \n";
                    printf "\t"; print_term (listfold_term_term_var_fp t1' def); printf "\n";
                    printf "\t"; print_term (listfold_term_term_var_fp t2' def); printf "\n";
                    printf ")\n\n";
                    None
            )
      )

    | (Match (t1,t2)) -> (
        
        match_infer t1 t2 0 0 ctxt def exec
          
      )
        
    | (MatchT (t1,t2,t3)) -> (
        
        Some t3
          
      )

and match_infer t1 t2 n m ctxt def exec =
  (*printf "match_infer (%d, %d)\n\n" n m;  *)
  match (infer t1 ctxt def exec) with
    | None -> None
    | Some t1' -> 
        let (t1'_fct, t1'_args) = (app_first_term_args t1') in 
          (
            match t1'_fct with 

              | (Ind(x,t11,t12)) -> (
                  match (nth t12 n, nth t2 m) with
                    | (Some c, Some d) -> (

                        let (con, fcv) = (forall_concl_list_var_type c) in 
                        let (i, iargs) = (app_first_term_args con) in 
                          
                        let (_, lvt) = (lambda_concl_list_var_type d (List.length fcv)) in 
                        let ctxt' = (rev(lvt) @ ctxt) in
                          
                        let ctxt'' = (rewrite_context_term_term ctxt' t1 (listargs_app (Cons (n,Ind(x,t11,t12))) (list_name2list_var (list_proj1 lvt)))) in
                        let hdd = (rewrite_term_term_term d t1 (listargs_app (Cons (n,Ind(x,t11,t12))) (list_name2list_var (list_proj1 lvt)))) in                                
                        let (hdd', _) = (lambda_concl_list_var_type hdd (List.length fcv)) in 
                          match (unification_list 
                                    (listrewrite_listterm_var_term iargs def)
                                    (listrewrite_listterm_var_term t1'_args  def)
                                    VarSet.empty) with
                            | NoMgu -> (
                                (*printf "\n\nno unification \n";*)
                                match_infer t1 t2 (n+1) m ctxt def exec
                              )
                            | Mgu s ->
                                
                                let s' = (match_comp_args (comp_subs s) fcv lvt) in
                                let ctxt''' = (apply_subst_context ctxt'' s' (*VarSet.empty*)(context_to_setvar lvt)) in
                                let hdd'' = (apply_subs_term hdd' s' (*VarSet.empty*)(context_to_setvar lvt)) in
                                  
                                  (infer hdd'' ctxt''' def exec)

                            | DnkMgu ->
                                
                                (infer hdd' ctxt'' def exec)
                      )
                    | _ -> None
                )
                  (* pas sur --> False ??? *)
              | t1' ->  (
                  match (listrewrite_term_var_term t1' (rev def)) with
                    | (Ind (x,t11,t12)) -> (

                        match (nth t12 n, nth t2 m) with
                          | (Some c, Some d) -> (
                              
                              let (con, fcv) = (forall_concl_list_var_type c) in 
                              let (i, iargs) = (app_first_term_args con) in 
                                
                              let (_, lvt) = (lambda_concl_list_var_type d (List.length fcv)) in 
                              let ctxt' = (rev(lvt) @ ctxt) in
                                
                              let ctxt'' = (rewrite_context_term_term ctxt' t1 (listargs_app (Cons (n,Ind(x,t11,t12))) (list_name2list_var (list_proj1 lvt)))) in
                              let hdd = (rewrite_term_term_term d t1 (listargs_app (Cons (n,Ind(x,t11,t12))) (list_name2list_var (list_proj1 lvt)))) in                                
                              let (hdd', _) = (lambda_concl_list_var_type hdd (List.length fcv)) in 
                                match (unification_list 
                                          (listrewrite_listterm_var_term iargs def)
                                          (listrewrite_listterm_var_term t1'_args  def)
                                          VarSet.empty) with
                                  | NoMgu -> (
                                      (*printf "\n\nno unification \n";*)
                                      match_infer t1 t2 (n+1) m ctxt def exec
                                    )
                                  | Mgu s ->
                                      
                                      let s' = (match_comp_args (comp_subs s) fcv lvt) in
                                      let ctxt''' = (apply_subst_context ctxt'' s' (*VarSet.empty*)(context_to_setvar lvt)) in
                                      let hdd'' = (apply_subs_term hdd' s' (*VarSet.empty*)(context_to_setvar lvt)) in
                                        
                                        (infer hdd'' ctxt''' def exec)

                                  | DnkMgu ->
                                      
                                      (infer hdd' ctxt'' def exec)
                                        
                            )
                          | _ -> None
                              
                              
                      )
                        
                    | t1'' -> (
                        printf "infer(match 3)\n\n";
                        print_term t1; printf "\n"; 
                        print_term t1'; printf "\n";
                        print_term t1''; printf "\n\n";
                        None
                      )
                        
                )
                  
          )
            
;;   


let beta_red_infer t ctxt def = 
  match (infer t ctxt def beta_red) with
    | None -> None
    | Some t' -> Some (beta_red t' (context_to_setvar ctxt) def);;

let simpl_infer t ctxt def = 
  match (infer t ctxt def simpl_red) with
    | None -> None
    | Some t' -> Some (simpl_red t' (context_to_setvar ctxt) def);;


(* Type checking  (def is only for pretty printing) *)

let rec typechecking te ty ctxt def =
  (*printf "("; print_context (listfold_context_term_var_fp ctxt def); printf ") |-\n\t"; print_term (listfold_term_term_var_fp te def); printf " : \n"; print_term (listfold_term_term_var_fp ty def); printf "\n\n"; *)
  (*printf "("; print_context (listfold_context_term_var_fp ctxt def); printf ") |-\n\t"; print_term te; printf " : \n"; print_term ty; printf "\n\n";*)
  match (te, ty) with
      
    | (Var x, _) -> (
        match (var_type_in_context x ctxt) with
          | None -> (
              printf "Var: false\n\n"; 
              printf "("; print_context (listfold_context_term_var_fp ctxt def); printf ") |-\n\t"; print_term (listfold_term_term_var_fp te def); printf " : \n"; print_term (listfold_term_term_var_fp ty def); printf "\n\n";
              false)
          | Some ty' -> (term_equiv ty' ty (context_to_setvar ctxt) def)
      )

    | (Type, Type) -> (true)

    | (Forall (x, ta, tb), Type) -> 
        if ((typechecking ta Type ctxt def) && 
               (typechecking tb Type ((x,ta)::ctxt) def)) then true else
          (printf "Forall: false\n\n"; false)

    | (Lambda (x,ta1,ta2), Forall (y,tb1,tb2)) ->
        if (term_equiv ta1 tb1 (context_to_setvar ctxt) def) then
          if (typechecking ta1 Type ctxt def) then
            if (typechecking ta2 
                   (subst_var_term tb2 y (Var x) (context_to_setvar ctxt))
                   ((x,ta1)::ctxt) def) then true 
            else
              (

                printf "----------------------------------------------\n";
                printf "Lambda1: false\n\n"; 
                printf "("; 
                print_context (listfold_context_term_var_fp ((x,ta1)::ctxt) def); 
                printf ") |-\n\t"; 
                print_term (listfold_term_term_var_fp ta2 def); 
                printf " : \n"; 
                print_term (listfold_term_term_var_fp (subst_var_term tb2 y (Var x) (context_to_setvar ctxt)) def); 
                printf "\n";                               
                printf "----------------------------------------------\n\n";
               false)
          else
            (printf "Lambda2: false\n\n"; false)
        else 
          (
            printf "Lambda3: false\n"; 
            print_term ta1; printf "\n";
            print_term tb1; printf "\n\n";
            false
          )

    | (Ind (x,ta1,ta2), _) ->
        if (term_equiv ta1 ty (context_to_setvar ctxt) def) then (
          if ((typechecking ta1 Type ctxt def) && 
                 (list_typechecking ta2 (built_n_elt_list Type (List.length ta2)) ((x,ta1)::ctxt) def)) then true else (printf "Ind: false(2)\n\n"; false)
        ) else (printf "Ind: false\n\n"; false)

    | (Phi (x, n, ta1, ta2), _) ->
        let (type_concl, type_lcv) = (forall_nth_concl_list_var_type ta1 n) in
          if ((typechecking ta1 Type ctxt def) && 
                 (typechecking 
                     (concl_list_var_type_lambda
                         ta2
                         type_lcv
                     ) ty 
                     ((x,ta1)::ctxt) 
                     def
                 )
          ) then true else
            (printf "Phi: false\n\n"; false)

    | (Cons (n, (Ind (x, ta1, ta2))), _) -> (
        match (nth ta2 n) with
          | None -> 
              (printf "Cons: false\n\n"; false)
          | Some t -> (
              if 
                ((term_equiv 
                     (*
                       (cf_term_listvar 
                       (rewrite_term_var_term 
                       t 
                       x 
                       (Ind (x, ta1, ta2))
                       ) (context_to_setvar ctxt)) *)
                     (subst_var_term t x (Ind (x, ta1, ta2)) (context_to_setvar ctxt))
                     ty) (x ++ context_to_setvar ctxt) def && 
                    (typechecking (Ind (x, ta1, ta2)) ta1 ctxt def)) then true else
                (printf "Cons: false\n\n"; false)
            )
      )

    | (Cons (n, t1'), _) -> (
        match (listrewrite_term_var_term t1' (rev def)) with
          | (Ind (x,t1,t2)) -> (

              typechecking (Cons (n,Ind (x,t1,t2))) ty ctxt def;
              
            )
              
          | _ -> (
              printf "typechecking(cons 3)\n\n"; 
              false
            )
              
      )

    | (App (t1, t2), _) -> (
        match (beta_red_infer t1 ctxt def, beta_red_infer t2 ctxt def) with
          | (Some (Forall (x1, t11, t12)), Some t2') ->
              if (term_equiv t11 t2' (context_to_setvar ctxt) def) then
                if ((typechecking t1 (Forall (x1, t11, t12)) ctxt def)) then
                  if (typechecking t2 t2' ctxt def) then
                    if (term_equiv 
                           (*
                             (cf_term_listvar 
                             (rewrite_term_var_term 
                             t12 
                             x1 
                             t2
                             ) 
                             (x1 ++ (context_to_setvar ctxt))
                             )
                           *)
                           (subst_var_term t12 x1 t2 (context_to_setvar ctxt))
                           ty (x1 ++ (context_to_setvar ctxt))) def then true 
                    else
                      (printf "typechecking app(5)\n\n"; 
                       printf "("; print_context (listfold_context_term_var_fp ctxt def); printf ") |-\n\t"; print_term (listfold_term_term_var_fp te def); printf " : \n"; print_term (listfold_term_term_var_fp ty def); printf "\n\n";
                       
                       print_term (listfold_term_term_var_fp 
                                      (subst_var_term t12 x1 t2 (context_to_setvar ctxt))
                                      def
                       ); 
                       printf " =/= \n"; 
                       print_term (listfold_term_term_var_fp ty def); printf "\n\n";
                       false) 
                  else
                    (printf "typechecking app(6)\n\n"; 
                     printf "("; print_context (listfold_context_term_var_fp ctxt def); printf ") |-\n"; 
                     printf "\t"; print_term t2; printf "\n";
                     printf " : \n"; print_term t2'; printf "\n\n";                
                     false)
                else
                  (printf "typechecking app(7)\n\n"; false)
              else (
                printf "typechecking app(1)\n\n";
                
                printf "("; print_context (listfold_context_term_var_fp ctxt def); printf ") |-\n"; 
                printf "App(\n";
                printf "\t"; print_term (listfold_term_term_var_fp t1 def); printf "\n";
                printf "\t"; print_term (listfold_term_term_var_fp t2 def); printf "\n";
                printf ")";
                printf " : \n"; print_term (listfold_term_term_var_fp ty def); printf "\n\n";                
                
                print_term (listfold_term_term_var_fp t11 def); printf " =/= \n";
                print_term (listfold_term_term_var_fp t2' def); printf "\n\n";                    

                printf "\n\n"; false
              )
          | (None, _) -> (
              printf "typechecking app(2)\n\n"; 
              printf "("; print_context (listfold_context_term_var_fp ctxt def); printf ") |-\n"; 
              printf "App(\n";
              printf "\t"; print_term (listfold_term_term_var_fp t1 def); printf "\n";
              printf "\t"; print_term (listfold_term_term_var_fp t2 def); printf "\n";
              printf ")";
              printf " : \n"; print_term (listfold_term_term_var_fp ty def); printf "\n\n";                
              false
            )
          | (Some t1', None) -> (
              printf "typechecking app(3)\n\n"; 
              printf "("; print_context (listfold_context_term_var_fp ctxt def); printf ") |-\n"; 
              printf "App(\n";
              printf "\t"; print_term (listfold_term_term_var_fp t1 def); printf "\n";
              printf "\t"; print_term (listfold_term_term_var_fp t2 def); printf "\n";
              printf ")";
              printf " : \n"; print_term (listfold_term_term_var_fp ty def); printf "\n\n";                
              
              print_term (listfold_term_term_var_fp t1' def); printf "\n";

              
              false)
          | (Some t1', Some t2') -> (
              match (simpl_red (listrewrite_term_var_term t1' def) VarSet.empty def) with
                | (Forall (x1, t11, t12)) ->
                    if (term_equiv t11 t2' (context_to_setvar ctxt) def &&
                           (typechecking t1 (Forall (x1, t11, t12)) ctxt def) &&
                           (typechecking t2 t2' ctxt def) &&
                           (term_equiv                                 
                               (subst_var_term t12 x1 t2 (context_to_setvar ctxt))
                               ty (x1 ++ (context_to_setvar ctxt)) def)) then true else
                      (printf "typechecking app(4)\n\n"; false)
                | _ ->                            
                    printf "typechecking app(5)\n\n"; 
                    printf "("; print_context (listfold_context_term_var_fp ctxt def); printf ") |-\n"; 
                    printf "App(\n";
                    printf "\t"; print_term (listfold_term_term_var_fp t1 def); printf "\n";
                    printf "\t"; print_term (listfold_term_term_var_fp t2 def); printf "\n";
                    printf ")";
                    printf " : \n"; print_term (listfold_term_term_var_fp ty def); printf "\n\n";                
                    
                    false
            )

      )

    | (Match (t1, t2), _) -> (
        match (beta_red_infer t1 ctxt def) with
          | None -> false
          | Some t1' -> (
              (* printf "type of Matched: "; print_term (listfold_term_term_var_fp t1' def); printf "\n\n"; *)
              let (t1'', args) = (app_first_term_args t1') in
                match (t1'') with
                  | Ind (x,t11,t12) -> (

                      (match_typechecking 0 t12 t2 x (Ind (x,t11,t12)) args t1 ty ctxt def)

                    )
                  | _ -> (
                      match (beta_red_infer t1 ctxt def) with
                        | None -> false
                        | Some t1' -> (
                            
                            let (t1'', args) = (app_first_term_args (simpl_red (listrewrite_term_var_term t1' def) VarSet.empty def)) in
                              match (listrewrite_term_var_term t1'' (rev def)) with
                                | Ind (x,t11,t12) -> (
                                    
                                    (match_typechecking 0 t12 t2 x (Ind (x,t11,t12)) args t1 ty ctxt def)
                                      
                                  )
                                | t1'' -> (
                                    printf "not an inductive type\n"; 
                                    print_term t1; printf "\n";
                                    print_term t1'; printf "\n";
                                    print_term t1''; printf "\n\n";
                                    false
                                  )
                          )
                    )
            )
      )

    | (MatchT (t1, t2, t3), _) -> (

        if (typechecking (Match (t1,t2)) ty ctxt def) then true else
          (printf "typechecking MatchT\n\n"; false)

      )

    | (_, App (t1, t2)) -> 
        
        let (fct, args) = (app_first_term_args ty) in
        let fct' = (match fct with
          | Var x -> unfold_fct fct x def
          | _ -> fct
        ) in
        let ty' = (simpl_red (listargs_app fct' args) VarSet.empty def) in
          if (ty' = ty) then
            false else
            typechecking te ty' ctxt def

    | _ -> false

and list_typechecking lte lty ctxt def =
  (*printf "list_typechecking\n";*)
  match (lte, lty) with
    | ([],[]) -> true
    | (lte1::lte2, lty1::lty2) ->
        if (typechecking lte1 lty1 ctxt def) then
          (list_typechecking lte2 lty2 ctxt def) else false
    | _ -> false

and match_typechecking n lc ld x ind args t ty ctxt def =
  (*printf "match_typechecking (%d, %d)\n\n" (List.length lc) (List.length ld); *)
  match lc with
    | [] -> (
        match ld with
          | [] -> true
          | _ -> (printf "match (%d,%d)\n\n" (List.length lc) (List.length ld); false)
      )              
    | (hdc::tlc) -> (
        if (not (valid_destructor hdc x ind args ctxt def)) then (
          (match_typechecking (n+1) tlc ld x ind args t ty ctxt def)
        )
        else
          match ld with
            | [] -> (
                printf "match DnkMgu (%d)\n" n;               
                (printf "match (%d,%d)\n\n" (List.length lc) (List.length ld); false)
              )
            | (hdd::tld) -> (
                
                match (destruct_deptype n hdc hdd t x args ind ty ctxt def) with
                  | None -> (
                      if (match_typechecking (n+1) tlc ld x ind args t ty ctxt def) then
                        true
                      else
                        false
                    )
                  | Some (
                      (te', ty', ctxt'),
                      (te'', ty'', ctxt'')
                    ) ->
                      if (
                        (typechecking te' ty' ctxt' def) &&
                          (typechecking te'' ty'' ctxt'' def)
                      ) then (
                        
                        (match_typechecking (n+1) tlc tld x ind args t ty ctxt def)
                      ) else (
                        
                        printf "*****************************************\n\n"; 
                        printf "ctxt' = "; print_context (listfold_context_term_var_fp ctxt' def); printf"\n"; 
                        printf "te' = "; print_term (listfold_term_term_var_fp te' def); printf"\n"; 
                        printf "ty' = "; print_term (listfold_term_term_var_fp ty' def); printf"\n"; 
                        printf "--------------------------------------------\n\n"; 
                        printf "ctxt'' = "; print_context (listfold_context_term_var_fp ctxt'' def); printf"\n"; 
                        printf "te'' = "; print_term (listfold_term_term_var_fp te'' def); printf"\n"; 
                        printf "ty'' = "; print_term (listfold_term_term_var_fp ty'' def); printf"\n"; 
                        printf "*****************************************\n\n";                       
                        false
                      )
              )
      )
;;

let alpha_typechecking te ty ctxt def =
  let (ctxt',l1) = (cf_context' ctxt VarSet.empty) in 
  let (te', l2) = (cf_term' te l1) in
  let (ty', l3) = (cf_term' ty l2) in (
    typechecking te' ty' ctxt' def
  );;


let rec valid_context' c ctxt def =
  match c with
    | [] -> true
    | (hd1, hd2) :: tl ->
        if (typechecking hd2 Type ctxt def) then
          (valid_context' tl ((hd1, hd2)::ctxt) def) else false;;

let valid_context c def = valid_context' (rev c) [] def;;

