(*
 This file is part of Poussin.

 Poussin is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Poussin is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Poussin.  If not, see <http://www.gnu.org/licenses/>.

 Copyright (C) 2008 Nicolas Marti
*)

open List;;
open String;;
open Printf;;

open Def;;
open Pprinter;;
open Listutil;;
open Freshness;;
open Term;;
open Varset;;

(*******************)
(* a term equality *)
(*******************)

let rec term_eq' t1 t2 l =
  
  match (t1, t2) with

    | (Var x1, Var x2) ->
        if (x1 = x2) then (true, l) else (

          (*printf "alpha_rel(var): %s =/= %s\n\n" x1 x2;*)
          (false, l)

        )

    | (Type, Type) -> (true, l)

    | (Lambda (x1, t11,t12), Lambda (x2, t21, t22)) ->
        if (x1 = x2) then (
          match (term_eq' t11 t21 (x1 ++ l)) with
            | (true, l1) -> (
                match (term_eq' t12 t22 l1) with
                  | (true, l2) ->
                      (true, l2)
                  | (false, l2) ->
                      ((false, l2))
              )
            | (false, l1) -> ((false, l1))
        ) else (
          let x = (fresh_name_list_name (l +++ (term_set_var (Lambda (x1, t11,t12))) +++ (term_set_var (Lambda (x2, t21,t22)))) x1) in
          match (cf_term' (rewrite_term_var_term t11 x1 (Var x)) (x ++ l), cf_term' (rewrite_term_var_term t21 x2 (Var x)) (x ++ l)) with
            | ((t11', l1'),(t21', l2')) ->
                match (cf_term' (rewrite_term_var_term t12 x1 (Var x)) l1', cf_term' (rewrite_term_var_term t22 x2 (Var x)) l2') with
                  | ((t12', l1''),(t22', l2'')) ->
                      match (term_eq' t11' t21' (x ++ l)) with
                        | (true, l1) -> (
                            match (term_eq' t12' t22' l1) with
                              | (true, l2) ->
                                  (true, l2)
                              | (false, l2) ->
                                  ((false, l2))
                          )
                        | (false, l1) -> ((false, l1))
                            
        ) 

    | (Forall (x1, t11,t12), Forall (x2, t21, t22)) ->
        if (x1 = x2) then (
          match (term_eq' t11 t21 (x1 ++ l)) with
            | (true, l1) -> (
                match (term_eq' t12 t22 l1) with
                  | (true, l2) ->
                      (true, l2)
                  | (false, l2) ->
                      ((false, l2))
              )
            | (false, l1) -> ((false, l1))
        ) else (
          let x = (fresh_name_list_name (l +++ (term_set_var (Forall (x1, t11,t12))) +++ (term_set_var (Forall (x2, t21,t22)))) x1) in
          match (cf_term' (rewrite_term_var_term t11 x1 (Var x)) (x ++ l), cf_term' (rewrite_term_var_term t21 x2 (Var x)) (x ++ l)) with
            | ((t11', l1'),(t21', l2')) ->
                match (cf_term' (rewrite_term_var_term t12 x1 (Var x)) l1', cf_term' (rewrite_term_var_term t22 x2 (Var x)) l2') with
                  | ((t12', l1''),(t22', l2'')) ->
                      match (term_eq' t11' t21' (x ++ l)) with
                        | (true, l1) -> (
                            match (term_eq' t12' t22' l1) with
                              | (true, l2) ->
                                  (true, l2)
                              | (false, l2) ->
                                  (false, l2)
                          )
                        | (false, l1) -> (false, l1)               
                            
        ) 

    | (Phi (x1, n1, t11,t12), Phi (x2, n2, t21, t22)) ->
        if (x1 = x2) then (
          let (t11_concl, t11_lcv) = (forall_concl_list_var_type t11) in
          let (t21_concl, t21_lcv) = (forall_concl_list_var_type t21) in
            term_eq' 
              
              (concl_list_var_type_lambda
                  (App (t11_concl, t12))
                  t11_lcv
              )

              (concl_list_var_type_lambda
                  (App (t21_concl, t22))
                  t21_lcv
              )
              
              (x1 ++ (x2 ++ l))
            
            
        ) else (
          (* Here is the probleme !!!! *)
          let l' = (l +++ (term_set_var (Phi (x1, n1, t11,t12))) +++ (term_set_var (Phi (x2, n2, t21,t22)))) in
          let x = (fresh_name_list_name l' x1) in
          let t12' = (rewrite_term_var_term t12 x1 (Var x)) in
          let t22' = (rewrite_term_var_term t22 x2 (Var x)) in
          let (t11_concl, t11_lcv) = (forall_concl_list_var_type t11) in
          let (t21_concl, t21_lcv) = (forall_concl_list_var_type t21) in
            match (term_eq' 
                      
                      (concl_list_var_type_lambda
                          (App (t11_concl, t12'))
                          t11_lcv
                      )
                      
                      (concl_list_var_type_lambda
                          (App (t21_concl, t22'))
                          t21_lcv
                      )
                      
                      (x1 ++ (x ++ (x2 ++ l')))
            ) with
              | (false, l'') ->
                  (false, l'')
              | t'' -> t''
                            
        ) 

    | (Ind (x1, t11,t12), Ind (x2, t21, t22)) ->
        if (x1 = x2) then (
          match (term_eq' t11 t21 (x1 ++ l)) with
            | (true, l1) -> (
                match (term_eq_list' t12 t22 l1) with
                  | (true, l2) ->
                      (true, l2)
                  | (false, l2) ->
                      ((false, l2))
              )
            | (false, l1) -> ((false, l1))
        ) else (
          let x = (fresh_name_list_name (l +++ (term_set_var (Ind (x1, t11,t12))) +++ (term_set_var (Ind (x2, t21,t22)))) x1) in
          match (cf_term' (rewrite_term_var_term t11 x1 (Var x)) (x ++ l), cf_term' (rewrite_term_var_term t21 x2 (Var x)) (x ++ l)) with
            | ((t11', l1'),(t21', l2')) ->
                match (cf_listterm' (rewrite_listterm_var_term t12 x1 (Var x)) l1', cf_listterm' (rewrite_listterm_var_term t22 x2 (Var x)) l2') with
                  | ((t12', l1''),(t22', l2'')) ->
                      match (term_eq' t11' t21' (x ++ l)) with
                        | (true, l1) -> (
                            match (term_eq_list' t12' t22' l1) with
                              | (true, l2) ->
                                  (true, l2)
                              | (false, l2) ->
                                  ((false, l2))
                          )
                        | (false, l1) -> ((false, l1))
                            
        ) 

    | (Cons (n1, t11), Cons (n2, t21)) ->
        if (n1 = n2) then (term_eq' t11 t21 l) else (false, l)

    | (App (t11, t12), App (t21, t22)) -> (
        match (term_eq' t11 t21 l) with
          | (true, l1) -> (
              match (term_eq' t12 t22 l1) with
                | (true, l2) ->
                    (true, l2)
                | (false, l2) ->
                    ((*printf "alpha_rel(app(1))\n\n";*) false, l2)
            )
          | (false, l1) -> (((*printf "alpha_rel(app(2))\n\n"; *) false, l1))        
      )

    | (Match (t11, t12), Match (t21, t22)) -> (
        match (term_eq' t11 t21 l) with
          | (true, l1) -> (
              match (term_eq_list' t12 t22 l1) with
                | (true, l2) ->
                    (true, l2)
                | (false, l2) ->
                    ((false, l2))
            )
          | (false, l1) -> ((false, l1))
      )

    | (MatchT (t11, t12, t13), MatchT (t21, t22, t23)) -> (
        
        (term_eq' (Match (t11, t12)) (Match (t21, t22)) l)
        
      )

    | _ -> (
        (*
          printf "term_eq\n";
          printf "t1 = "; print_term t1; printf "\n";
          printf "t2 = "; print_term t2; printf "\n\n";
          printf "muri(term_eq) \n\n";
        *)
        (false, l)
      )

and term_eq_list' t1 t2 l =  
  (*printf "alpha_rel_list'\n\n"; *)
  match (t1, t2) with
    | ([], []) -> 
        ((true, l))
    | (hd1::tl1, hd2::tl2) -> (
        match (term_eq' hd1 hd2 l) with
          | (true, l1) ->              
              (term_eq_list' tl1 tl2 l)
          | (false, _) -> 
              (false, l)                
      )
    | _ -> ((*printf "muri(alpha_rel_list')\n";*) (false, l))

;;

let term_eq t1 t2 = 
  let (b, l) = (term_eq' t1 t2 VarSet.empty) in b;;

(* fold term *)

let rec fold_term_term_var t x y =

  if (term_eq t y) then (Var x) else (
    
    match t with
      | Lambda (s, t1,t2) -> (          
          Lambda (s,(fold_term_term_var t1 x y),(fold_term_term_var t2 x y))          
        )

      | Forall (s, t1,t2) -> (
          Forall (s,(fold_term_term_var t1 x y),(fold_term_term_var t2 x y))          
        )

      | Ind (s, t1,t2) -> (
          Ind (s,(fold_term_term_var t1 x y),(fold_listterm_term_var t2 x y))          
        )

      | Phi (s, n, t1,t2) -> (
          Phi (s, n, (fold_term_term_var t1 x y),(fold_term_term_var t2 x y))
        )

      | Cons (n, t1) -> (
          Cons (n,(fold_term_term_var t1 x y))
        )

      | App (t1,t2) -> (
          App ((fold_term_term_var t1 x y),(fold_term_term_var t2 x y))
        )

      | Match (t1, t2) -> (
          Match ((fold_term_term_var t1 x y),(fold_listterm_term_var t2 x y))
        )

      | MatchT (t1, t2, t3) -> (
          MatchT ((fold_term_term_var t1 x y),(fold_listterm_term_var t2 x y),(fold_term_term_var t3 x y))
        )
      | _ -> t
  )

and fold_listterm_term_var l x y =
  
  match l with 
    | [] -> []
    | hd :: tl ->
        (fold_term_term_var hd x y) :: (fold_listterm_term_var tl x y);;


let rec listfold_term_term_var t l =
  match l with
    | [] -> t 
    | (hd1, hd2) :: tl ->
        listfold_term_term_var (fold_term_term_var t hd1 hd2) tl;;

let rec listfold_term_term_var_fp t l =
  let t' = (listfold_term_term_var t l) in
    if (t = t') then t else (listfold_term_term_var_fp t' l);;

let rec listfold_listterm_term_var_fp t l =
  match t with
    | [] -> []
    | hd :: tl ->
        (listfold_term_term_var_fp hd l) :: (listfold_listterm_term_var_fp tl l);;

let rec listfold_context_term_var_fp t l =
  match t with
    | [] -> []
    | (hd1, hd2) :: tl ->
        (hd1, (listfold_term_term_var_fp hd2 l)) :: (listfold_context_term_var_fp tl l);;


  let rec readable_term t def =
    match def with
      | [] -> t
      | hd :: tl ->
          readable_term (listfold_term_term_var_fp t hd) tl;;

  let rec readable_term_fp t def =
  let t' = (readable_term t def) in
    if (t = t') then t else (readable_term_fp t' def);;

  let rec readable_context_fp c def =
    match c with
      | [] -> []
      | (hd1, hd2) :: tl ->
          (hd1, readable_term_fp hd2 def) :: (readable_context_fp tl def);;

let rec unfold_fct t n def =
  match def with
    | [] -> t
    | (hd1,hd2) :: tl ->
        if (hd1 = n) then
          rewrite_term_var_term t n hd2
        else
          unfold_fct t n tl
;;
